"""Trading module."""
import logging

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from alpaca.common import APIError
from alpaca.trading import Order, TradeAccount, TradingClient
from alpaca.trading.requests import GetOrdersRequest, QueryOrderStatus
from typeguard import typechecked

from optitrader.config import SETTINGS
from optitrader.market.market_data import MarketData
from optitrader.portfolio import Portfolio

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class AlpacaTrading(TradingClient):
    """Class to interact with Alpaca trading API."""

    def __init__(
        self,
        api_key: str | None = None,
        secret_key: str | None = None,
        paper: bool = True,
    ) -> None:
        super().__init__(
            api_key or SETTINGS.ALPACA_TRADING_API_KEY,
            secret_key=secret_key or SETTINGS.ALPACA_TRADING_API_SECRET,
            paper=paper,
        )
        self.market_data = MarketData(
            trading_key=api_key or SETTINGS.ALPACA_TRADING_API_KEY,
            trading_secret=secret_key or SETTINGS.ALPACA_TRADING_API_SECRET,
        )

    @property
    def account(self) -> TradeAccount:
        """Get the account."""
        account = self.get_account()
        assert isinstance(account, TradeAccount)
        return account

    def get_portfolio(self) -> Portfolio:
        """Get Portfolio from starting positions."""
        pos = self.get_all_positions()
        assert isinstance(pos, list)
        _pos_series = pd.Series({p.symbol: float(p.market_value) for p in pos})
        return Portfolio(weights=_pos_series / _pos_series.sum(), market_data=self.market_data)

    def get_account_portfolio_history(self) -> pd.DataFrame:
        """
        Get the account portfolio history from Alpaca.

        See more: https://alpaca.markets/docs/api-references/trading-api/portfolio-history/
        """
        df = pd.DataFrame(self.get("/account/portfolio/history?timeframe=1D"))
        _idx = "timestamp"
        df[_idx] = pd.to_datetime(df[_idx], unit="s").dt.date
        df.set_index(_idx, inplace=True)
        df.drop(columns=["base_value", "timeframe"], inplace=True)
        return df

    def get_account_portfolio_history_plot(self) -> go.Figure:
        """Get the plot of the account portfolio history."""
        history = self.get_account_portfolio_history()["equity"]
        return px.line(
            data_frame=history,
            x=history.index,
            y=history.values,
            labels={"timestamp": "", "y": ""},
            title="",
        )

    def get_orders_df(self, status: QueryOrderStatus = QueryOrderStatus.ALL) -> pd.DataFrame:
        """Get the account orders in a df."""
        orders = self.get_orders(
            GetOrdersRequest(
                limit=500,  # current max
                status=status,
            )
        )
        assert isinstance(orders, list)
        _idx = "symbol"
        df = pd.DataFrame([o.dict() for o in orders])
        if _idx in df.columns:
            return df.set_index(["symbol"])
        return df

    @typechecked
    def invest_in_portfolio(
        self, portfolio: Portfolio, amount: float, fail_on_error: bool = False
    ) -> list[Order]:
        """Invest a certain amount in a portfolio."""
        bp = self.account.buying_power
        assert bp, "No buying power in the account."
        _bp = round(number=float(bp), ndigits=2)
        amount = round(number=float(amount), ndigits=2)
        assert (
            _bp > amount
        ), f"The passed `amount` {amount} is greater then the account buying power {_bp}"
        acct = self.get_account()
        assert isinstance(acct, TradeAccount)
        orders_resp = []
        for order in portfolio.to_orders_list(amount=amount):
            try:
                resp = self.submit_order(order)
                assert isinstance(resp, Order)
                orders_resp.append(resp)
            except APIError as alpaca_api_error:
                log.warning(order.symbol)
                log.warning(order.notional)
                log.warning(acct.buying_power)
                log.warning(alpaca_api_error)
                if fail_on_error:
                    raise AssertionError(
                        f"Order to {order.side} {order.notional} $ of {order.symbol} failed!"
                    ) from alpaca_api_error
        return orders_resp



"""Init."""
from optitrader.utils.utils import clean_string, rearrange_columns_by_zeros, remove_punctuation

__all__ = [
    "clean_string",
    "remove_punctuation",
    "rearrange_columns_by_zeros",
]



"""Objective function module."""
from abc import ABCMeta, abstractmethod

import cvxpy as cp
import numpy as np
import pandas as pd
from pydantic import BaseModel

from optitrader.enums import ObjectiveName


class _BaseObjectiveModel(BaseModel):
    """Base objective model to avoid duplication."""

    name: ObjectiveName


class ObjectiveValue(_BaseObjectiveModel):
    """Model to represent the mapping of an objective to its optimal value."""

    value: float
    weight: float


class OptimizationVariables(_BaseObjectiveModel):
    """Objective optimization variables."""

    minimize: cp.Minimize

    class Config:
        """Configuration."""

        arbitrary_types_allowed = True


class PortfolioObjective(metaclass=ABCMeta):
    """Objective function abstract class."""

    def __init__(
        self,
        name: ObjectiveName,
        weight: float = 1.0,
    ) -> None:
        self.weight = weight
        self.name = name

    @abstractmethod
    def get_objective_and_auxiliary_constraints(
        self,
        returns: pd.DataFrame,
        weights_variable: cp.Variable,
    ) -> tuple[OptimizationVariables, list[cp.Constraint]]:
        """Get optimization matrices."""

    @abstractmethod
    def get_obj_latex(self) -> str:
        """Get objective formulation as latex."""


class CVaRObjectiveFunction(PortfolioObjective):
    """
    Conditional value at Risk objective function.

    Conditional Value at Risk (CVaR), also known as Expected Shortfall (ES),
    is a risk measure used in financial analysis and portfolio management
    to assess the potential losses beyond a specified confidence level.
    Unlike traditional risk measures such as Value at Risk (VaR)
    that focus solely on the magnitude of potential losses,
    CVaR provides additional insights by considering the severity of losses
    that occur beyond the VaR threshold.
    """

    def __init__(
        self,
        confidence_level: float = 0.7,
        weight: float = 1.0,
    ) -> None:
        super().__init__(weight=weight, name=ObjectiveName.CVAR)
        self.confidence_level = confidence_level

    def get_objective_and_auxiliary_constraints(
        self,
        returns: pd.DataFrame,
        weights_variable: cp.Variable,
    ) -> tuple[OptimizationVariables, list[cp.Constraint]]:
        """Get CVaR optimization matrices."""
        rets_vals = returns.values
        n_obs = rets_vals.shape[0]
        losses_minus_var = cp.Variable(n_obs, nonneg=True)
        value_at_risk = cp.Variable(1)
        objective_function = value_at_risk + 1 / ((1 - self.confidence_level) * n_obs) * cp.sum(
            losses_minus_var
        )
        return (
            OptimizationVariables(
                name=self.name, minimize=cp.Minimize(self.weight * objective_function)
            ),
            [-rets_vals @ weights_variable - value_at_risk - losses_minus_var <= 0],
        )

    def get_obj_latex(self) -> str:
        """Get objective formulation as latex."""
        return r"""
        \text{CVaR}_{\alpha}(X) = \frac{1}{1 - \alpha} \int_{\alpha}^{1} \text{VaR}_{\beta}(X) \, d\beta
        """


class CovarianceObjectiveFunction(PortfolioObjective):
    """Variance objective function."""

    def __init__(
        self,
        weight: float = 1.0,
    ) -> None:
        super().__init__(weight=weight, name=ObjectiveName.COVARIANCE)

    def get_objective_and_auxiliary_constraints(
        self,
        returns: pd.DataFrame,
        weights_variable: cp.Variable,
    ) -> tuple[OptimizationVariables, list[cp.Constraint]]:
        """Get Variance optimization matrices."""
        # Annualize the cov mat
        sigma = 252 * returns.cov().values
        objective_function = weights_variable @ sigma @ weights_variable
        return (
            OptimizationVariables(
                name=self.name, minimize=cp.Minimize(self.weight * objective_function)
            ),
            [],
        )

    def get_obj_latex(self) -> str:
        """Get objective formulation as latex."""
        return r"""
        \text{Var}(X) = \frac{1}{n} \sum_{i=1}^{n} (X_i - \bar{X})^2
        """


class ExpectedReturnsObjectiveFunction(PortfolioObjective):
    """Expected Returns objective function."""

    def __init__(
        self,
        weight: float = 0.25,
    ) -> None:
        super().__init__(weight=weight, name=ObjectiveName.EXPECTED_RETURNS)

    def get_objective_and_auxiliary_constraints(
        self,
        returns: pd.DataFrame,
        weights_variable: cp.Variable,
    ) -> tuple[OptimizationVariables, list[cp.Constraint]]:
        """Get Expected Returns optimization matrices."""
        # Annualize expected returns and put minus in front to maximize
        exp_rets = -252 * returns.mean().values
        objective_function = weights_variable @ exp_rets
        return (
            OptimizationVariables(
                name=self.name, minimize=cp.Minimize(self.weight * objective_function)
            ),
            [],
        )

    def get_obj_latex(self) -> str:
        """Get objective formulation as latex."""
        return r"""
        \text{E}(R_p) = \sum_{i=1}^{n} w_i \times \text{E}(R_i)
        """


class MADObjectiveFunction(PortfolioObjective):
    """
    ### Mean Absolute Deviation objective function.

    Mean Absolute Deviation measures the average deviation of data points from the mean. It is a risk measure that focuses on the dispersion of the data.

    The formula for Mean Absolute Deviation (MAD) is given by:
    """

    def __init__(
        self,
        weight: float = 1.0,
    ) -> None:
        super().__init__(weight=weight, name=ObjectiveName.MEAN_ABSOLUTE_DEVIATION)

    def get_objective_and_auxiliary_constraints(
        self,
        returns: pd.DataFrame,
        weights_variable: cp.Variable,
    ) -> tuple[OptimizationVariables, list[cp.Constraint]]:
        """Get Mean Absolute Deviation optimization matrices."""
        rets_vals = returns.values
        n_obs = rets_vals.shape[0]
        abs_devs = cp.Variable(n_obs, nonneg=True)
        rets_minus_mu = rets_vals - returns.mean().values
        objective_function = cp.sum(abs_devs) / n_obs
        return OptimizationVariables(
            name=self.name, minimize=cp.Minimize(self.weight * objective_function)
        ), [
            rets_minus_mu @ weights_variable - abs_devs <= 0,
            -rets_minus_mu @ weights_variable - abs_devs <= 0,
        ]

    def get_obj_latex(self) -> str:
        """Get objective formulation as latex."""
        return r"""
        \text{MAD}(X) = \frac{1}{n} \sum_{i=1}^{n} |X_i - \bar{X}|
        """


class FinancialsObjectiveFunction(PortfolioObjective):
    """
    Financials objective function.

    This objective lets you maximize one item on the financial statements, such as the net profit.
    """

    def __init__(
        self,
        weight: float = 0.25,
    ) -> None:
        super().__init__(weight=weight, name=ObjectiveName.FINANCIALS)

    def get_objective_and_auxiliary_constraints(
        self,
        returns: pd.DataFrame,
        weights_variable: cp.Variable,
    ) -> tuple[OptimizationVariables, list[cp.Constraint]]:
        """
        Get Financials optimization matrices.

        Note that the returns are actually the financials.
        """
        # Boost the financials growth and put minus in front to maximize
        objective_function = cp.sum(
            weights_variable @ -returns * 1e-4
        )  # 1e-4 is the scaling factor to compare with other objectives
        return (
            OptimizationVariables(
                name=self.name, minimize=cp.Minimize(self.weight * objective_function)
            ),
            [],
        )

    def get_obj_latex(self) -> str:
        """Get objective formulation as latex."""
        return r"""
        \text{E}(R_p) = \sum_{i=1}^{n} w_i \times R_{i,scaled}
        """


class MostDiversifiedObjectiveFunction(PortfolioObjective):
    """
    ### Most diversified portfolio (MDP) objective function.

    This objective lets you maximize the diversification ratio, that is
    the ratio of the weighted average of volatilities divided by the portfolio volatility.

    #### Mathematical Formulation:

    Let `n` be the number of assets in the portfolio.
    Let `returns` be a 2D numpy array of shape (T, n), where `T` is the number of time periods
    and `n` is the number of assets.`C` is the covariance matrix of asset returns with shape (n, n).
    `w` is a vector of portfolio weights with lenght `n`.
    `sigma` is the portfolio volatility.
    'vol' is a vector of asset volatilities with shape (n,), obtained from the diagonal elements of 'C'.
    'sum' represents the sum of elements in a vector.

    The diversification ratio (`DR`) can be formulated as follows:
    """

    def __init__(
        self,
        weight: float = 1.0,
    ) -> None:
        super().__init__(weight=weight, name=ObjectiveName.MOST_DIVERSIFIED)

    def get_objective_and_auxiliary_constraints(
        self,
        returns: pd.DataFrame,
        weights_variable: cp.Variable,
    ) -> tuple[OptimizationVariables, list[cp.Constraint]]:
        """
        Get Financials optimization matrices.

        Note that the returns are actually the financials.
        """
        # copy the weights to not modify the other weights variable
        _weights = cp.Variable(weights_variable.shape[0], nonneg=True)
        cov_matrix = returns.cov().values
        return (
            OptimizationVariables(
                name=self.name,
                minimize=cp.Minimize(_weights @ cov_matrix @ _weights),
            ),
            [_weights @ np.sqrt(cov_matrix.diagonal()) == 1],
        )

    def get_obj_latex(self) -> str:
        """Get objective formulation as latex."""
        return r"""
        DR = \frac{\sum{w * vol}}{\sqrt{w^T * C * w}}
        """


class ObjectivesMap:
    """Objectives map."""

    def __init__(
        self,
        objectives: list[PortfolioObjective] | None = None,
    ) -> None:
        self.objective_mapping: dict[ObjectiveName, type[PortfolioObjective]] = {
            ObjectiveName.CVAR: CVaRObjectiveFunction,
            ObjectiveName.EXPECTED_RETURNS: ExpectedReturnsObjectiveFunction,
            ObjectiveName.MEAN_ABSOLUTE_DEVIATION: MADObjectiveFunction,
            ObjectiveName.COVARIANCE: CovarianceObjectiveFunction,
            ObjectiveName.FINANCIALS: FinancialsObjectiveFunction,
            ObjectiveName.MOST_DIVERSIFIED: MostDiversifiedObjectiveFunction,
        }
        self.objectives: list[PortfolioObjective] = objectives or []

    @property
    def objectives_names(self) -> list[str]:
        """Return the obejctives names."""
        return [o.name for o in self.objectives]

    def to_objective(
        self,
        name: ObjectiveName,
        weight: float = 1.0,
    ) -> PortfolioObjective:
        """Get an objective."""
        return self.objective_mapping[name](weight=weight)  # type: ignore

    def reset_objectives_names(self, objectives_names: list[ObjectiveName]) -> None:
        """Reset the objective names."""
        self.objectives = [self.to_objective(name) for name in objectives_names]

    def get_objective_by_name(
        self,
        name: ObjectiveName,
    ) -> PortfolioObjective:
        """Get an objective."""
        _objs = [o for o in self.objectives if o.name == name]
        assert len(_objs) > 0, f"Objective {name} not in the objectives list {self.objectives}."
        return _objs[0]

    def add_objective(
        self,
        name: ObjectiveName,
        weight: float = 1.0,
    ) -> None:
        """Add an objective to the map."""
        if name not in self.objectives_names:
            if name == ObjectiveName.FINANCIALS:
                self.objectives.append(
                    self.to_objective(
                        name=name,
                        weight=weight,
                    )
                )
            else:
                self.objectives.append(self.to_objective(name=name, weight=weight))
        else:
            obj = self.get_objective_by_name(name)
            if obj.weight != weight:
                obj.weight = weight

    def get_obj_doc(
        self,
        name: ObjectiveName,
    ) -> str:
        """Return the objective docstring."""
        return (
            self.objective_mapping[name].__doc__
            or f"{self.objective_mapping[name].__name__} documentation."
        )

    def get_obj_latex(
        self,
        name: ObjectiveName,
    ) -> str:
        """Return the objective docstring."""
        return self.objective_mapping[name]().get_obj_latex()  # type: ignore



"""Streamlit about page."""

import logging

from optitrader.app import Page, session

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def main() -> None:
    """Run dashboard."""
    page = Page(name="Alpaca account dashboard")
    page.display_title_and_description(
        description="""
        Boost the portfolio analytics for your Alpaca account.
        """
    )
    session.display_trader_portfolio()
    session.display_alpaca_account_sidebar()
    page.display_code_sidebar()



"""Backtester enums."""
from optitrader.enums.iterable import IterEnum


class RebalanceFrequency(IterEnum):
    """Supported rebalance frequencies."""

    WEEKLY = "W"
    MONTHLY = "M"
    QUARTERLY = "3M"
    YEARLY = "12M"



"""Module to handle market data."""

import asyncio
import logging
import time
from functools import lru_cache

import finnhub
import pandas as pd

from optitrader.config import SETTINGS
from optitrader.enums import BarsField, DataProvider
from optitrader.enums.market import BalanceSheetItem, CashFlowItem, IncomeStatementItem
from optitrader.market.alpaca_market_data import AlpacaMarketData, Asset
from optitrader.market.base_data_provider import BaseDataProvider
from optitrader.market.db.database import MarketDB
from optitrader.market.finnhub_market_data import FinnhubClient
from optitrader.market.news import NewsArticle
from optitrader.market.yahoo_market_data import YahooMarketData
from optitrader.models.asset import AssetModel, _YahooFinnhubCommon

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class MarketData:
    """Class that implements market data connections."""

    def __init__(
        self,
        data_provider: DataProvider = DataProvider.ALPACA,
        trading_key: str | None = SETTINGS.ALPACA_TRADING_API_KEY,
        trading_secret: str | None = SETTINGS.ALPACA_TRADING_API_SECRET,
        broker_key: str | None = SETTINGS.ALPACA_BROKER_API_KEY,
        broker_secret: str | None = SETTINGS.ALPACA_BROKER_API_SECRET,
        use_db: bool = True,
    ) -> None:
        self._trading_key = trading_key
        self._trading_secret = trading_secret
        self._broker_key = broker_key
        self._broker_secret = broker_secret
        self.__alpaca_client = AlpacaMarketData(
            trading_key=trading_key,
            trading_secret=trading_secret,
            broker_key=broker_key,
            broker_secret=broker_secret,
        )
        self.__yahoo_client = YahooMarketData()
        self.__finnhub = FinnhubClient() if SETTINGS.FINHUB_API_KEY else None
        provider_mapping: dict[DataProvider, BaseDataProvider] = {
            DataProvider.ALPACA: self.__alpaca_client,
            DataProvider.YAHOO: self.__yahoo_client,
        }
        self.__provider_client = provider_mapping[data_provider]
        self.use_db = use_db
        if self.use_db:
            self._db = MarketDB()

    @lru_cache  # noqa: B019
    def load_prices(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
        bars_field: BarsField = BarsField.CLOSE,
    ) -> pd.DataFrame:
        """
        Load the prices df from the data provider.

        Parameters
        ----------
        `tickers`: tuple[str, ...]
            A tuple of str representing the tickers.
        `start_date`: pd.Timestamp
            A pd.Timestamp representing start date.
        `end_date`: pd.Timestamp
            A pd.Timestamp representing end date.
        `bars_field`: BarsField
            A field in the OHLCV bars. Defaults to CLOSE.
        `excel_filename` str | None
            The path for an excel file with custom prices.
            Defaults to None and is required only if the DataProvider is EXCEL_FILE.

        Returns
        -------
        `prices`
            pd.DataFrame with market prices.
        """
        return self.__provider_client.get_prices(
            tickers=tickers,
            start_date=start_date,
            end_date=end_date,
            bars_field=bars_field,
        )

    def get_total_returns(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
        required_pct_obs: float = 0.95,
    ) -> pd.DataFrame:
        """
        Return total return dataframe.

        Parameters
        ----------
        `tickers`: tuple[str, ...]
            A tuple of str representing the tickers.
        `start_date`: pd.Timestamp
            A pd.Timestamp representing start date.
        `end_date`: pd.Timestamp
            A pd.Timestamp representing end date.
        `required_pct_obs`: float
            Minimum treshold for non NaNs in each column.
            Columns with more NaNs(%)>required_pct_obs will be dropped.

        Returns
        -------
        `returns`
            pd.DataFrame with market linear returns.
        """
        returns = (
            self.load_prices(
                tickers=tickers,
                start_date=start_date,
                end_date=end_date,
            )
            .pct_change()
            .iloc[1:, :]
        )
        # remove tickers that do not have enough observations
        return returns.dropna(axis=1, thresh=int(returns.shape[0] * required_pct_obs))

    def get_asset(self, ticker: str) -> AssetModel:
        """
        Return asset info from ticker.

        Parameters
        ----------
        `ticker`: str
            A str representing the ticker.

        Returns
        -------
        `asset`
            AssetModel data model.
        """
        if self.use_db:
            return self._db.get_asset(ticker)
        return self.get_asset_from_ticker(ticker)

    @lru_cache  # noqa: B019
    def get_asset_from_ticker(self, ticker: str) -> AssetModel:
        """
        Return asset info from ticker.

        Parameters
        ----------
        `ticker`: str
            A str representing the ticker.

        Returns
        -------
        `asset`
            AssetModel data model.
        """
        duplicate_fields = set(_YahooFinnhubCommon.__fields__.keys())
        apca_asset = self.__alpaca_client.get_alpaca_asset(ticker)
        finnhub_asset = self.__finnhub.get_asset_profile(ticker) if self.__finnhub else None
        yahoo_asset = self.__yahoo_client.get_yahoo_asset(ticker)
        return AssetModel(
            **apca_asset.dict(exclude_none=True),
            **yahoo_asset.dict(
                exclude=duplicate_fields,
                exclude_none=True,
            )
            if yahoo_asset
            else {},
            **finnhub_asset.dict(
                exclude=duplicate_fields,
                exclude_none=True,
            )
            if finnhub_asset
            else {},
        )

    def _get_asset_from_ticker(self, ticker: str) -> AssetModel | None:
        """
        Return asset info from ticker.

        Parameters
        ----------
        `ticker`: str
            A str representing the ticker.

        Returns
        -------
        `asset`
            AssetModel data model.
        """
        try:
            return self.get_asset_from_ticker(ticker)
        except Exception as error:
            log.debug(ticker)
            log.debug(type(error))
            return None

    async def _async_get_assets(self, tickers: tuple[str, ...]) -> list[AssetModel]:
        """
        Return assets info from ticker.

        Parameters
        ----------
        `tickers`: tuple(str)
            A tuple of str representing the tickers.

        Returns
        -------
        `assets`
            A list of AssetModel data model.
        """
        _threads = [asyncio.to_thread(self._get_asset_from_ticker, ticker) for ticker in tickers]
        return await asyncio.gather(*_threads)

    @lru_cache  # noqa: B019
    def get_assets_from_provider(self, tickers: tuple[str, ...]) -> list[AssetModel]:
        """
        Return assets info from ticker.

        Parameters
        ----------
        `tickers`: tuple(str)
            A tuple of str representing the tickers.

        Returns
        -------
        `assets`
            A list of AssetModel data model.
        """
        tickers_bucket_size = 30
        assets = []
        start = time.time()
        for i in range(0, len(tickers), tickers_bucket_size):
            tickers_bucket = tickers[i : i + tickers_bucket_size]
            try:
                _new_assets = asyncio.run(self._async_get_assets(tickers=tickers_bucket))
            except finnhub.FinnhubAPIException as api_error:
                reset_remaining = 60 - (time.time() - start)
                log.warning(f"Request for tickers {tickers_bucket} sleeping {reset_remaining}")
                log.warning(type(api_error))
                if reset_remaining:
                    time.sleep(reset_remaining)  # wait time limit reset
                _new_assets = asyncio.run(self._async_get_assets(tickers=tickers_bucket))
            assets.extend([a for a in _new_assets if a is not None])
        return assets

    def get_assets(self, tickers: tuple[str, ...] | None = None) -> list[AssetModel]:
        """
        Return assets info from tickers.

        If self.use_db it returns from the database,
        otherwise from the data provider.

        Parameters
        ----------
        `tickers`: tuple(str)
            A tuple of str representing the tickers.

        Returns
        -------
        `assets`
            A list of AssetModel data model.
        """
        if self.use_db:
            return self._db.get_asset_models(tickers)
        assert tickers, "Use the tickers or set use_db = True"
        return self.get_assets_from_provider(tickers)

    def get_assets_df(
        self,
        tickers: tuple[str, ...] | None = None,
    ) -> pd.DataFrame:
        """
        Return assets info from ticker.

        Parameters
        ----------
        `tickers`: tuple(str)
            A tuple of str representing the tickers.

        Returns
        -------
        `assets`
            A list of AssetModel data model.
        """
        if self.use_db:
            return self._db.get_assets_df(tickers)
        return pd.DataFrame([a.dict() for a in self.get_assets(tickers)])

    @lru_cache  # noqa: B019
    def get_financials(self, ticker: str) -> pd.DataFrame:
        """
        Return asset info from ticker.

        Parameters
        ----------
        `ticker`: str
            A str representing the ticker.

        Returns
        -------
        `fin_df`
            pd.DataFrame of financials.
        """
        return self.__yahoo_client.get_financials(ticker)

    @lru_cache  # noqa: B019
    def get_multi_financials_by_item(
        self,
        tickers: tuple[str, ...],
        financial_item: IncomeStatementItem
        | CashFlowItem
        | BalanceSheetItem = IncomeStatementItem.NET_INCOME,
    ) -> pd.DataFrame:
        """
        Return asset info from tickers.

        Parameters
        ----------
        `tickers`: str
            A str representing the tickers.
        `financial_item`: IncomeStatementItem
        | CashFlowItem
        | BalanceSheetItem = IncomeStatementItem.NET_INCOME

        Returns
        -------
        `fin_df`
            pd.DataFrame of financials.
        """
        return self.__yahoo_client.get_multi_financials_by_item(
            tickers, financial_item=financial_item
        )

    def get_tradable_tickers(self) -> tuple[str, ...]:
        """Get all tradable tickers from Alpaca."""
        return tuple(self.__alpaca_client.get_alpaca_tickers())

    def get_asset_by_name(self, name: str) -> Asset:
        """Get asset by name from Alpaca."""
        return self.__alpaca_client.get_asset_by_name(name)

    def get_total_number_of_shares(
        self,
        tickers: tuple[str, ...],
    ) -> pd.Series:
        """Get the number of shares for each ticket in tickets."""
        if self.use_db:
            return self._db.get_number_of_shares(tickers).set_index("ticker")
        return self.__yahoo_client.get_multi_number_of_shares(tickers)

    def get_market_caps(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        """
        Return total return dataframe.

        Parameters
        ----------
        `tickers`: tuple[str, ...]
            A tuple of str representing the tickers.
        `start_date`: pd.Timestamp
            A pd.Timestamp representing start date.
        `end_date`: pd.Timestamp
            A pd.Timestamp representing end date.
        `required_pct_obs`: float
            Minimum treshold for non NaNs in each column.
            Columns with more NaNs(%)>required_pct_obs will be dropped.

        Returns
        -------
        `returns`
            pd.DataFrame with market linear returns.
        """
        return self.load_prices(
            tickers=tickers,
            start_date=start_date,
            end_date=end_date,
        ).mul(self.get_total_number_of_shares(tickers)["number_of_shares"], axis=1)

    def get_top_market_caps(
        self,
        tickers: tuple[str, ...],
        top: int,
    ) -> pd.Series:
        """Get the tickers with the top market cap."""
        caps = self.get_market_caps(
            tickers=tickers,
            # only taking last week data for the market cap
            start_date=pd.Timestamp.today() - pd.Timedelta(days=5),
        )
        return caps.iloc[-1, :].sort_values(ascending=False)[:top]

    def get_top_market_cap_tickers(
        self,
        tickers: tuple[str, ...],
        top: int = 10,
    ) -> tuple[str, ...]:
        """Get the tickers with the top market cap."""
        return tuple(self.get_top_market_caps(tickers=tickers, top=top).index)

    def get_news(
        self,
        tickers: tuple[str, ...],
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        limit: int = 5,
        include_content: bool = True,
        exclude_contentless: bool = True,
    ) -> list[NewsArticle]:
        """Get news articles."""
        return self.__alpaca_client.get_news(
            tickers=tickers,
            start=start,
            end=end,
            limit=limit,
            include_content=include_content,
            exclude_contentless=exclude_contentless,
        )

    def get_news_df(
        self,
        tickers: tuple[str, ...],
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        limit: int = 5,
        include_content: bool = True,
        exclude_contentless: bool = True,
    ) -> pd.DataFrame:
        """Get news articles in a df."""
        return self.__alpaca_client.get_news_df(
            tickers=tickers,
            start=start,
            end=end,
            limit=limit,
            include_content=include_content,
            exclude_contentless=exclude_contentless,
        )



"""Local SQL storage of assets table."""
import logging
import os
from datetime import datetime

import pandas as pd
from sqlalchemy import create_engine, select, update
from sqlalchemy.exc import DatabaseError
from sqlalchemy.orm import Session, sessionmaker

from optitrader.config import SETTINGS
from optitrader.market.db.models import Asset, Base
from optitrader.models.asset import AssetModel

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class MarketDB:
    """Class to handle interactions with sqlite market.db database."""

    def __init__(self, uri: str = SETTINGS.DB_URI_MARKET) -> None:
        """Initialize the market database object."""
        self.engine = create_engine(uri, connect_args={"check_same_thread": False})
        self._SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        self.session: Session = self._SessionLocal()

    def create_tables(self) -> None:
        """Create the assets table."""
        with self.engine.begin() as conn:
            Base.metadata.create_all(conn)

    def drop_tables(self) -> None:
        """Drop the assets table."""
        with self.engine.begin() as conn:
            Base.metadata.drop_all(conn)

    def get_assets(
        self,
        tickers: tuple[str, ...] | None = None,
    ) -> list[Asset]:
        """Get all the assets in the table."""
        if tickers:
            return list(
                self.session.execute(select(Asset).filter(Asset.ticker.in_(tickers)))
                .scalars()
                .fetchall()
            )
        return list(self.session.execute(select(Asset)).scalars().fetchall())

    def get_asset(self, ticker: str) -> AssetModel:
        """Get the asset model from the table by ticker."""
        return AssetModel.from_orm(
            list(
                self.session.execute(select(Asset).where(Asset.ticker == ticker))
                .scalars()
                .fetchall()
            )[0]
        )

    def get_asset_models(
        self,
        tickers: tuple[str, ...] | None = None,
    ) -> list[AssetModel]:
        """Get all the assets in the table."""
        return [AssetModel.from_orm(a) for a in self.get_assets(tickers)]

    def get_tickers(self) -> list[str]:
        """Get all the tickers in the assets table."""
        return list(
            self.session.execute(select(Asset.ticker).order_by(Asset._id)).scalars().fetchall()
        )

    def get_assets_df(
        self,
        tickers: tuple[str, ...] | None = None,
    ) -> pd.DataFrame:
        """Get all the tickers in the assets table."""
        query = select(Asset).filter(Asset.ticker.in_(tickers)) if tickers else select(Asset)
        with self.engine.begin() as conn:
            return pd.read_sql_query(sql=query, con=conn)

    def get_number_of_shares(
        self,
        tickers: tuple[str, ...] | None = None,
    ) -> pd.DataFrame:
        """Get all the tickers in the assets table."""
        query = (
            select(Asset.ticker, Asset.number_of_shares).filter(Asset.ticker.in_(tickers))
            if tickers
            else select(Asset.ticker, Asset.number_of_shares)
        )
        with self.engine.begin() as conn:
            return pd.read_sql_query(sql=query, con=conn)

    def update_number_of_shares(
        self,
        multi_number_of_shares: pd.Series,
        updated_by: str | None = None,
    ) -> pd.DataFrame:
        """Update the number_of_shares in the assets table."""
        updated_by = updated_by or os.path.relpath(__file__)
        tickers = multi_number_of_shares.index
        for ticker in tickers:
            num_shares = int(multi_number_of_shares.loc[ticker])
            if num_shares > 0:
                query = (
                    update(Asset)
                    .where(Asset.ticker == ticker)
                    .values(
                        number_of_shares=num_shares,
                        updated_at=datetime.utcnow(),
                        updated_by=updated_by,
                    )
                )
                with self.engine.begin() as conn:
                    conn.execute(query)
        return self.get_number_of_shares(tickers=tuple(tickers))

    def write_asset(
        self,
        asset_model: AssetModel,
        updated_by: str | None = None,
        autocommit: bool = True,
    ) -> None:
        """Write assets in the database."""
        asset = Asset(
            updated_by=updated_by,
            **asset_model.dict(exclude_none=True, exclude={"symbol"}),
        )
        self.session.add(asset)
        if autocommit:
            self.session.commit()
            log.info(f"Added {asset}.")

    def write_assets(
        self,
        asset_models: list[AssetModel],
        updated_by: str | None = None,
        autocommit: bool = True,
    ) -> None:
        """Write assets in the database."""
        for asset_model in asset_models:
            try:
                self.write_asset(asset_model, updated_by=updated_by, autocommit=autocommit)
            except DatabaseError as dberror:
                log.warning(f"{type(dberror)} for {asset_model.ticker}")
                log.warning(dberror)
                if autocommit:
                    self.session.rollback()
        log.info(f"Added {len(asset_models)} assets.")



"""optitrader REST API."""

import logging
from collections import OrderedDict

import coloredlogs
import pandas as pd
from fastapi import FastAPI, HTTPException, status
from fastapi.responses import RedirectResponse

from optitrader.market.investment_universe import InvestmentUniverse
from optitrader.market.market_data import MarketData
from optitrader.models import OptimizationRequest, OptimizationResponse
from optitrader.optimization.solver import Solver

app = FastAPI(title="optitrader API")


@app.on_event("startup")
def startup_event() -> None:
    """Run API startup events."""
    # Remove all handlers associated with the root logger object.
    for handler in logging.root.handlers:
        logging.root.removeHandler(handler)
    # Add coloredlogs' coloured StreamHandler to the root logger.
    coloredlogs.install()


@app.get("/")
def reroute_to_docs() -> RedirectResponse:
    """Automatically redirect homepage to docs."""
    return RedirectResponse(url="/docs")


@app.post("/optimization")
def compute_optimal_portfolio(
    request_body: OptimizationRequest,
) -> OptimizationResponse:
    """Compute the optimal portfolio."""
    market = MarketData()
    if not (request_body.tickers or request_body.universe_name):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="You must provide either tickers or universe name",
        )
    opt_ptf = Solver(
        returns=market.get_total_returns(
            tickers=InvestmentUniverse(name=request_body.universe_name).tickers
            if request_body.universe_name
            else request_body.tickers
            if request_body.tickers  # for mypy
            else (),
            start_date=pd.Timestamp(request_body.start_date),
            end_date=pd.Timestamp(request_body.end_date),
        ),
        constraints=[con.to_ptf_constraint() for con in request_body.constraints],
        objectives=[obj.to_ptf_objective() for obj in request_body.objectives],
    ).solve(
        weights_tolerance=request_body.weights_tolerance,
    )
    return OptimizationResponse(
        weights=opt_ptf.get_non_zero_weights().to_dict(OrderedDict),
        objective_values=opt_ptf.objective_values,
    )



"""Base models for the optimization API router."""


from collections import OrderedDict
from datetime import date, datetime, timedelta

from optitrader.config import SETTINGS
from optitrader.enums import ConstraintName
from optitrader.market.investment_universe import UniverseName
from optitrader.models.base import CustomBaseModel as BaseModel
from optitrader.optimization.constraints import (
    ConstraintsMap,
    PortfolioConstraint,
)
from optitrader.optimization.objectives import (
    ObjectiveName,
    ObjectivesMap,
    ObjectiveValue,
    PortfolioObjective,
)


class ObjectiveModel(BaseModel):
    """Objective model for the opt request."""

    _obj_map: ObjectivesMap = ObjectivesMap()
    name: ObjectiveName
    weight: float

    def to_ptf_objective(self) -> PortfolioObjective:
        """Parse to objective."""
        return self._obj_map.to_objective(name=self.name, weight=self.weight)


class ConstraintModel(BaseModel):
    """Constraint model for the opt request."""

    _constr_map: ConstraintsMap = ConstraintsMap()
    name: ConstraintName

    def to_ptf_constraint(self) -> PortfolioConstraint:
        """Parse to constraint."""
        return self._constr_map.to_constraint(self.name)


class OptimizationRequest(BaseModel):
    """Optimization request body."""

    tickers: tuple[str, ...] | None
    universe_name: UniverseName | None
    start_date: date = datetime.utcnow().date() - timedelta(days=365 * 2)
    end_date: date = datetime.utcnow().date() - timedelta(days=1)
    objectives: list[ObjectiveModel]
    constraints: list[ConstraintModel] = [  # noqa: RUF012
        ConstraintModel(name=ConstraintName.SUM_TO_ONE),
        ConstraintModel(name=ConstraintName.LONG_ONLY),
    ]
    weights_tolerance: float | None = SETTINGS.SUM_WEIGHTS_TOLERANCE


class OptimizationResponse(BaseModel):
    """Optimization response body."""

    weights: OrderedDict[str, float]
    objective_values: list[ObjectiveValue]



"""Investment Universe module."""
from functools import lru_cache

import pandas as pd
from pydantic import BaseModel

from optitrader.enums import UniverseName


class _WikiScrapeRequest(BaseModel):
    """Universe scraped info."""

    html_index: int
    url_path: str
    column_name: str


class InvestmentUniverse:
    """Class that implements an investment universe."""

    SCRAPED_UNIV = (UniverseName.NASDAQ, UniverseName.SP500)

    def __init__(
        self,
        tickers: tuple[str, ...] | None = None,
        name: UniverseName | None = None,
    ) -> None:
        assert tickers or name, "Only one of tickers or name must be provided."
        if tickers:
            self.tickers = tickers
        if name:
            self.name = name
            if name in self.SCRAPED_UNIV:
                self.tickers = self.scrape_wikipedia_tickers()
            else:
                univ_name_mapping: dict[UniverseName, tuple[str, ...]] = {
                    UniverseName.FAANG: (
                        "META",  # F
                        "AAPL",  # A
                        "AMZN",  # A
                        "NFLX",  # N
                        "GOOGL",  # G
                    ),
                    UniverseName.POPULAR_STOCKS: (
                        "AAPL",
                        "AMZN",
                        "TSLA",
                        "GOOGL",
                        "BRK.B",
                        "V",
                        "JPM",
                        "NVDA",
                        "MSFT",
                        "DIS",
                        "NFLX",
                        "META",
                        "WMT",
                        "BABA",
                        "AMD",
                        "ACN",
                        "PFE",
                        "ORCL",
                        "ZM",
                        "SHOP",
                        "COIN",
                    ),
                }
                self.tickers = univ_name_mapping[name]

    def __len__(self) -> int:
        """Length of universe tickers."""
        return len(self.tickers)

    @lru_cache  # noqa: B019
    def scrape_wikipedia_tickers(self) -> tuple[str, ...]:
        """Scrape wikixpedia.com for universe name tickers."""
        assert self.name, "The name must be set to use this method."
        assert self.name in self.SCRAPED_UNIV, f"The name must be one of {self.SCRAPED_UNIV}."
        _scraped_univ_map: dict[UniverseName, _WikiScrapeRequest] = {
            UniverseName.NASDAQ: _WikiScrapeRequest(
                html_index=4,
                url_path="Nasdaq-100",
                column_name="Ticker",
            ),
            UniverseName.SP500: _WikiScrapeRequest(
                html_index=0,
                url_path="List_of_S%26P_500_companies",
                column_name="Symbol",
            ),
        }
        params = _scraped_univ_map[self.name]
        # TODO: these tables have a lot of information such as changes, GICS sectors and industries and name
        _html = pd.read_html(f"https://en.wikipedia.org/wiki/{params.url_path}", flavor="html5lib")
        tickers: tuple[str, ...] = tuple(_html[params.html_index][params.column_name])
        # basic validation on the tickers
        for t in tickers:
            assert isinstance(t, str), f"Found ticker {t} that's not a string."
            assert t.isupper(), f"Found ticker {t} that's not upper case."
        return tickers



"""Streamlit about page."""

import logging

from optitrader.app import Page, session

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def main() -> None:
    """Run dashboard."""
    page = Page(name="Backtest your strategy")
    page.display_title_and_description(
        description="""
        The backtesting page is a user-friendly tool that allows traders and investors to assess the historical performance of their trading strategies.\n
        It offers a simple interface to check the optimization parameters and view basic metrics such as profitability and portfolio allocation over time.\n
        The dashboard presents clear visualizations, including equity curves and performance summaries, enabling users to gain insights into their strategy's effectiveness.\n
        👈 You can see the optimization problem in the _sidebar_ and the backtester is going to rebalance
        your optimal portfolio based on your _rebalance frequency_ to and evaluate the performance on unseen data. \n
        """
    )
    session.set_rebalance_frequency()
    session.run_backtest()
    session.display_backtest()
    session.display_optitrader_problem()



"""All optimization related Enums."""

from optitrader.enums.iterable import IterEnum


class ConstraintName(IterEnum):
    """Support constraints."""

    SUM_TO_ONE = "Weights must sum to one"
    LONG_ONLY = "No shortselling allowed"
    NUMER_OF_ASSETS = "Number of assets in portfolio"
    WEIGHTS_PCT = "Weights (%) in portfolio"

    @property
    def is_bounded(self) -> bool:
        """Return true if this objective name is for a bounded constraint."""
        return self.name in ["NUMER_OF_ASSETS", "WEIGHTS_PCT"]


class ObjectiveName(IterEnum):
    """Supported objectives names."""

    MOST_DIVERSIFIED = "Maximize Diversification Ratio"
    CVAR = "Conditional Value at Risk"
    COVARIANCE = "Covariance"
    EXPECTED_RETURNS = "Expected Return"
    MEAN_ABSOLUTE_DEVIATION = "Mean Absolute Deviation"
    FINANCIALS = "Financial Statements"



"""Module with data provider base class."""

from typing import Protocol

import pandas as pd

from optitrader.enums import BarsField


class BaseDataProvider(Protocol):
    """Abstract base class for a data provider."""

    def __init__(self) -> None:
        super().__init__()

    def get_bars(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        """Get bars from the data provider."""

    def get_prices(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
        bars_field: BarsField = BarsField.CLOSE,
    ) -> pd.DataFrame:
        """Get prices from the data provider."""



"""Streamlit base page component."""
import streamlit as st


class Page:
    """Page class."""

    def __init__(
        self,
        name: str,
    ) -> None:
        self.name = name
        st.set_page_config(
            page_title="optitrader",
            page_icon="🛠️",
            layout="wide",
        )

    def display_title(self) -> None:
        """Display title."""
        st.title(self.name.title())

    def display_title_and_description(self, description: str) -> None:
        """Display title and description."""
        self.display_title()
        with st.expander("Description", expanded=True):
            st.markdown(description)

    @staticmethod
    def display_code_sidebar(with_divider: bool = True) -> None:
        """Display github link to code in sidebar."""
        with st.sidebar:
            if with_divider:
                st.divider()
            st.header("💻 Code")
            st.code("pip install optitrader")
            st.markdown(
                "[![Repo](https://badgen.net/badge/icon/GitHub?icon=github&label)](https://github.com/Ale-Cas/optitrader)"
            )



"""All market data related Enums."""
from optitrader.enums import IterEnum


class DataProvider(IterEnum):
    """Supported data providers for market data."""

    ALPACA = "ALPACA"
    YAHOO = "YAHOO"


class BarsField(IterEnum):
    """Possile fields in the bars from the data provider."""

    OPEN = "open"
    HIGH = "high"
    LOW = "low"
    CLOSE = "close"
    VOLUME = "volume"


class UniverseName(IterEnum):
    """Supported universe names."""

    FAANG = "FAANG Stocks (Facebook, Apple, Amazon, Google & Netflix)"
    POPULAR_STOCKS = "Most Popular Stocks"
    NASDAQ = "NASDAQ 100"
    SP500 = "Standard & Poor 500"


class IncomeStatementItem(IterEnum):
    """
    Supported income statement fields.

    See https://finance.yahoo.com/quote/AAPL/financials?p=AAPL
    """

    REVENUE = "TotalRevenue"
    # -
    COST = "CostOfRevenue"
    # =
    GROSS_PROFIT = "GrossProfit"
    # -
    OPERATING_EXPENSE = "OperatingExpense"
    # =
    OPERATING_INCOME = "OperatingIncome"

    # After taxes and other expenses
    NET_INCOME = "NetIncome"
    EBIT = "EBIT"
    # EBITDA = "EBITDA"
    EBITDA = "NormalizedEBITDA"


class BalanceSheetItem(IterEnum):
    """
    Supported balance sheet fields.

    See https://finance.yahoo.com/quote/AAPL/financials?p=AAPL
    """

    ASSETS = "TotalAssets"
    LIABILITIES = "TotalLiabilitiesNetMinorityInterest"
    DEBT = "TotalDebt"
    WORKING_CAPITAL = "WorkingCapital"


class CashFlowItem(IterEnum):
    """
    Supported cash flow fields.

    See https://finance.yahoo.com/quote/AAPL/financials?p=AAPL
    """

    FREE = "FreeCashFlow"
    OPERATING = "OperatingCashFlow"
    INVESTING = "InvestingCashFlow"
    FINANCING = "FinancingCashFlow"



"""optitrader CLI."""

import subprocess

import typer

app = typer.Typer()


@app.command()
def say(message: str = "") -> None:
    """Say a message."""
    typer.echo(message)


@app.command()
def dashboard(port: int = 8000, launch: bool = True, timeout: float | None = None) -> None:
    """Open the streamlit dashboard."""
    subprocess.run(["poe", "app", "--port", str(port)], timeout=timeout)
    if launch:
        typer.launch(url=f"http://localhost:{port}/")



"""Streamlit session manager."""

import logging
from functools import partial

import pandas as pd
import plotly.express as px
import streamlit as st
from alpaca.trading import TradeAccount

from optitrader import MarketData, Optitrader, Portfolio
from optitrader.backtester import Backtester, Portfolios
from optitrader.enums import (
    ConstraintName,
    IterEnum,
    ObjectiveName,
    RebalanceFrequency,
    UniverseName,
)
from optitrader.enums.market import BalanceSheetItem, CashFlowItem, IncomeStatementItem
from optitrader.market.investment_universe import InvestmentUniverse
from optitrader.market.trading import AlpacaTrading
from optitrader.optimization.constraints import ConstraintsMap
from optitrader.optimization.objectives import ObjectivesMap
from optitrader.utils.utils import remove_punctuation

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class SessionManager:
    """
    Streamlit session manager.

    Attributes
    ----------
    `ticker`: str =
    `market_data`: MarketData | None = None
    `universe_name` = UniverseName.FAANG
    `objective_names` = [ObjectiveName.CVAR]
    `constraint_names` = [ConstraintName.SUM_TO_ONE, ConstraintName.LONG_ONLY]
    `start_date` = pd.Timestamp.today() - pd.Timedelta(value=365 * 2, unit="day")
    `rebalance_frequency` = RebalanceFrequency.MONTHLY
    `_opt_ptf`: Portfolio | None = None
    `_ptfs`: list[Portfolio] | None = None
    `_backtest_history`: pd.Series | None = None
    """

    def __init__(self) -> None:
        """Initialize a Streamlit session object."""
        self.market_data: MarketData = MarketData()
        self.trader: AlpacaTrading = AlpacaTrading()
        self.universe_name = UniverseName.FAANG
        self.objective_names = [ObjectiveName.CVAR]
        self.obj_map = ObjectivesMap()
        self.constraint_names = [ConstraintName.SUM_TO_ONE, ConstraintName.LONG_ONLY]
        self.constr_map = ConstraintsMap(constraint_names=self.constraint_names)
        self.start_date = pd.Timestamp.today() - pd.Timedelta(value=365 * 2, unit="day")
        self.rebalance_frequency = RebalanceFrequency.MONTHLY
        self._opt_ptf: Portfolio | None = None
        self._ptfs: list[Portfolio] | None = None
        self._backtest_history: pd.Series | None = None
        self._ticker: str = "AAPL"
        self.tickers = InvestmentUniverse(name=self.universe_name).tickers

    def set_market_data(self, market_data: MarketData) -> None:
        """Set the market_data connection."""
        self.market_data = market_data

    @property
    def ticker(self) -> str:
        """Explore page ticker."""
        return self._ticker.upper()

    def set_ticker(self, ticker: str) -> None:
        """Set the ticker."""
        self._ticker = ticker

    def set_tickers(self, universe_name: str) -> None:
        """Get the universe tickers."""
        self.tickers = InvestmentUniverse(name=UniverseName(universe_name)).tickers

    def display_tickers(self) -> None:
        """Display the tickers in the universe."""
        tickers = sorted(self.tickers)
        tickers_per_row = 10
        len_tickers = len(tickers)
        with st.expander(
            f"{self.universe_name} universe tickers", expanded=len_tickers < tickers_per_row
        ):
            if len_tickers > tickers_per_row:
                for i in range(0, len_tickers, tickers_per_row):
                    row = tickers[i : i + tickers_per_row]
                    cols = st.columns(tickers_per_row)
                    for idx, col in enumerate(cols):
                        with col:
                            if idx < len(row):
                                st.code(row[idx])
            else:
                cols = st.columns(len(tickers))
                for idx, col in enumerate(cols):
                    with col:
                        st.code(tickers[idx])

    def _clean_opt_ptf(
        self,
    ) -> None:
        """Clean the optimal portfolio cache."""
        self._opt_ptf = None

    def _clean_backtest_results(self) -> None:
        """Clean the optimal portfolio cache."""
        self._ptfs = None
        self._backtest_history = None

    def _change_universe(self, universe_name: str) -> None:
        """Clean the optimal portfolio cache and reset the."""
        self._clean_opt_ptf()
        self._clean_backtest_results()
        self.set_tickers(universe_name)

    def _from_selectbox(
        self,
        label: str,
        options: type[IterEnum],
        value: str,
        is_value: bool = True,
    ) -> str:
        """Return the selected value from the selectbox."""
        if options == RebalanceFrequency:
            on_change = self._clean_backtest_results
        elif options == UniverseName:
            on_change = partial(self._change_universe, value)
        else:
            on_change = self._clean_opt_ptf
        sel = st.selectbox(
            label=label,
            options=options.get_values_list() if is_value else options.get_names_list(),
            index=options.get_index_of_value(value),
            on_change=on_change,
        )
        return sel or ""

    def _from_multiselect(
        self,
        label: str,
        options: type[IterEnum],
        default: list | None = None,
    ) -> list[str]:
        """Return a list from the multiselect."""
        _opts = options.get_values_list()
        return st.multiselect(
            label=label,
            options=_opts,
            default=default,
            on_change=self._clean_opt_ptf,
        )

    def set_api_keys(self) -> None:
        """Set the API keys."""
        with st.form(key="api-keys"):
            api_key = st.text_input("Enter your Alpaca Trading API key")
            secret_key = st.text_input("Enter your Alpaca Trading API secret")
            keys_submitted = st.form_submit_button("Login")
        if api_key and secret_key and keys_submitted:
            self.market_data = MarketData(
                trading_key=api_key,
                trading_secret=secret_key,
            )
            # self.trader = AlpacaTrading(
            #     api_key=api_key,
            #     secret_key=secret_key,
            #     paper=True,  # TODO: make the user choose
            # )

    def _holdings_to_st(self, holdings_df: pd.DataFrame) -> None:
        """Display holdings df in streamlit."""
        key_cols = [
            "logo",
            "ticker",
            "name",
            "weight_in_ptf",
        ]
        cols_to_ignore = ["updated_at", "updated_by", "id"]
        st.dataframe(
            holdings_df,
            column_order=[
                *key_cols,
                *(c for c in holdings_df.columns if c not in key_cols and c not in cols_to_ignore),
            ],
            column_config={
                "logo": st.column_config.ImageColumn("logo", help="Company Logo", width="small")
            },
            hide_index=False,
            use_container_width=True,
        )

    def _orders_to_st(self, orders_df: pd.DataFrame) -> None:
        """Display orders df in streamlit."""
        st.dataframe(
            orders_df,
            column_order=[
                "created_at",
                "side",
                "notional",
                "qty",
                "status",
                "type",
                "time_in_force",
                "filled_at",
                "filled_qty",
            ],
            column_config={
                "notional": st.column_config.NumberColumn(
                    "notional ($)",
                    help="The notional value of the order in USD",
                    min_value=0,
                    max_value=1e9,
                    step=0.01,
                    format="$%d",
                ),
                "created_at": st.column_config.DatetimeColumn(
                    "created",
                    format="D MMM YYYY, h:mm a",
                    step=1,
                ),
                "filled_at": st.column_config.DatetimeColumn(
                    "filled",
                    format="D MMM YYYY, h:mm a",
                    step=1,
                ),
            },
            use_container_width=True,
        )

    def display_trader_portfolio(self) -> None:
        """Display the user portfolio using Alpaca trading API positions."""
        _max_expanded_len = 20
        _monetary_precision = 2
        acct = self.trader.get_account()
        assert isinstance(acct, TradeAccount)
        bp = acct.buying_power
        assert bp
        st.subheader("Current Portfolio")
        st.text(f"Connected to account # {acct.account_number}")
        col1, col2 = st.columns(2)
        with col1:
            st.metric("💵 Buying power", value=round(float(bp), _monetary_precision))
        with col2:
            assert acct.equity is not None
            assert acct.last_equity is not None
            ac_equity = round(float(acct.equity), _monetary_precision)
            equity_change = round(
                ac_equity - round(float(acct.last_equity), _monetary_precision), _monetary_precision
            )
            st.metric("💵 Equity", value=ac_equity, delta=equity_change or None)
        ptf = self.trader.get_portfolio()
        holdings_df = ptf.get_holdings_df()
        if not holdings_df.empty:
            with col1:
                tab1, tab2 = st.tabs(
                    [
                        f"This portfolio past performance since {self.start_date.date()}",
                        "Your account performance so far",
                    ]
                )
                with tab1:
                    st.plotly_chart(
                        figure_or_data=ptf.history_plot(
                            start_date=self.start_date,
                            title="",
                        ),
                        use_container_width=True,
                    )
                with tab2:
                    st.plotly_chart(
                        self.trader.get_account_portfolio_history_plot(),
                        use_container_width=True,
                    )
            with col2:
                st.plotly_chart(
                    figure_or_data=ptf.pie_plot(title=""),
                    use_container_width=True,
                )
            with st.expander("Positions", expanded=len(holdings_df) < _max_expanded_len):
                self._holdings_to_st(holdings_df)
        orders_df = self.trader.get_orders_df()
        if not orders_df.empty:
            with st.expander("Orders", expanded=len(orders_df) < _max_expanded_len):
                self._orders_to_st(orders_df)

    def set_universe_name(self) -> None:
        """Set the universe name."""
        self.universe_name = UniverseName(
            self._from_selectbox(
                label="Choose the investment universe.",
                options=UniverseName,
                value=self.universe_name.value,
            )
        )

    def set_start_date(self) -> None:
        """Set the optimization start_date."""
        self.start_date = pd.Timestamp(
            st.date_input(
                label="Choose the first date used to retrieve prices for the optimization.",
                value=self.start_date,
            )
        )

    def set_objective_names(self) -> None:
        """Set the objectives name."""
        self.objective_names = [
            ObjectiveName(o)
            for o in self._from_multiselect(
                label="Choose one or more objective functions for your optimal portfolio.",
                options=ObjectiveName,
                default=self.objective_names,
            )
        ]
        self.obj_map.reset_objectives_names(self.objective_names)

    def set_constraint_names(self) -> None:
        """Set the constraints name."""
        self.constraint_names = [
            ConstraintName(o)
            for o in self._from_multiselect(
                label="Choose one or more constraints for your optimal portfolio.",
                options=ConstraintName,
                default=self.constraint_names,
            )
        ]
        self.constr_map.reset_constraint_names(self.constraint_names)

    def set_objectives(self) -> None:
        """Set the objectives from the obejctives name."""
        st.sidebar.subheader("🎯 Objectives")
        _min = 0.1
        for obj in self.objective_names:
            with st.expander(obj, expanded=False):
                st.write(self.obj_map.get_obj_doc(ObjectiveName(obj)))
                latex = self.obj_map.get_obj_latex(ObjectiveName(obj))
                if latex:
                    st.latex(latex)
                self.obj_map.add_objective(
                    name=obj,
                    weight=st.number_input(
                        f"Enter `{obj}` weight",
                        min_value=_min,
                        max_value=_min * 1e2,
                        step=_min,
                        value=_min,
                        key=obj,
                        on_change=self._clean_opt_ptf,
                    ),
                )

    def set_constraints(self) -> None:
        """Set the constraints from the obejctives name."""
        st.sidebar.subheader("🎛️ Constraints")
        for con in self.constraint_names:
            if con.is_bounded:
                with st.expander(con):
                    st.write(f"{con} docs")
                    lower_bound = int(
                        st.number_input(
                            f"`{con.title()}` minimum",
                            min_value=1,
                            max_value=10,
                            step=1,
                            value=2,
                            on_change=self._clean_opt_ptf,
                        )
                    )
                    upper_bound = int(
                        st.number_input(
                            f"`{con.title()}` maximum",
                            min_value=1,
                            max_value=100,
                            step=1,
                            value=50,
                            on_change=self._clean_opt_ptf,
                        )
                    )
                    self.constr_map.set_constraint_bounds(
                        name=con,
                        lower_bound=lower_bound,
                        upper_bound=upper_bound,
                    )
            else:
                with st.expander(con):
                    st.write(f"{con} documentation.")

    def set_rebalance_frequency(self) -> None:
        """Set the objectives name."""
        self.rebalance_frequency = RebalanceFrequency[
            self._from_selectbox(
                label="Choose the rebalance frequency for the backtest.",
                options=RebalanceFrequency,
                is_value=False,
                value=self.rebalance_frequency,
            )
        ]

    def get_optitrader(self) -> Optitrader:
        """Get the optitrader instance."""
        return Optitrader(
            objectives=self.obj_map.objectives,
            constraints=self.constr_map.constraints,
            universe_name=self.universe_name,
            market_data=self.market_data or MarketData(),
        )

    def display_solution(self) -> None:
        """Display the solution."""
        if self._opt_ptf:
            with st.spinner("Elaborating solution"):
                st.plotly_chart(
                    figure_or_data=self._opt_ptf.pie_plot(),
                    use_container_width=True,
                )
                holdings_df = self._opt_ptf.get_holdings_df()
                self._holdings_to_st(holdings_df)
                st.plotly_chart(
                    figure_or_data=self._opt_ptf.history_plot(start_date=self.start_date),
                    use_container_width=True,
                )
                st.markdown("##### Optimal objectives details")
                col1, col2, col3 = st.columns(3)
                for o in self._opt_ptf.objective_values:
                    with col1:
                        st.markdown(f"Name: `{o.name}`")
                    with col2:
                        st.markdown(f"Weight: `{round(o.weight, 1)}`")
                    with col3:
                        st.markdown(f"Optimal value: `{round(o.value, 9)}`")

    def _run_optimization(self) -> None:
        """Run the optitrader solve."""
        opt = self.get_optitrader()
        with st.spinner("Optimization in progress"):
            try:
                opt_ptf = opt.solve(
                    start_date=self.start_date,
                    weights_tolerance=1e-3,
                )
                opt_ptf.set_market_data(opt.market_data)
                self._opt_ptf = opt_ptf
                self.set_ticker(
                    opt_ptf.get_non_zero_weights().sort_values(ascending=False).index[0]
                )
            except AssertionError as er:
                log.warning(er)
                st.error("Couldn't find a solution!")

    def run_optimization(self) -> None:
        """Run the optitrader solve."""
        if st.button(label="COMPUTE OPTIMAL PORTFOLIO"):
            if self._opt_ptf:
                st.info(
                    "You already computed an optimal portfolio, change the parameters above to discard the solution and compute a new one."
                )
            else:
                self._run_optimization()

    def run_backtest(self) -> None:
        """Run the backtester."""
        if st.button(label="BACKTEST STRATEGY"):
            opt = self.get_optitrader()
            with st.spinner("Backtest in progress"):
                backtester = Backtester(
                    opt=opt,
                    start=self.start_date,
                    rebal_freq=self.rebalance_frequency,
                )
                self._ptfs = backtester.compute_portfolios()
                self._backtest_history = backtester.compute_history_values()

    def display_backtest(self) -> None:
        """Display the backtester result."""
        if self._backtest_history is not None and not self._backtest_history.empty:
            st.plotly_chart(
                figure_or_data=px.line(
                    data_frame=self._backtest_history,
                    x=self._backtest_history.index,
                    y=self._backtest_history.values,
                    labels={"timestamp": "", "y": ""},
                    title="Out of sample strategy wealth from start date to today",
                ),
                use_container_width=True,
            )
        if self._ptfs:
            st.write("Portfolios allocation over time")
            st.dataframe(
                Portfolios(self._ptfs)
                .to_df()
                .sort_index(ascending=False)
                .style.background_gradient(cmap="Blues", axis=1)
            )

    def display_financials(self) -> None:
        """Display the financials."""
        st.markdown("#### Financial statements over time")
        fin_df = self.market_data.get_financials(self.ticker)
        statements: dict[str, list[IterEnum]] = {
            "Income Statement": [
                IncomeStatementItem.REVENUE,
                IncomeStatementItem.NET_INCOME,
            ],
            "Balance Sheet": [BalanceSheetItem.ASSETS, BalanceSheetItem.LIABILITIES],
            "Cash Flow": [CashFlowItem.FREE, CashFlowItem.OPERATING],
        }
        tabs = st.tabs(list(statements.keys()))
        for _idx, tab in enumerate(tabs):
            statement = list(statements.keys())[_idx]
            fin_df_tab = fin_df[statements[statement]]
            with tab:
                st.plotly_chart(
                    figure_or_data=px.bar(
                        data_frame=fin_df_tab,
                        barmode="group",
                        labels={"asOfDate": "Date", "value": "💵   U.S. Dollars"},
                        title=statement,
                    ),
                    use_container_width=True,
                )

    def display_news(self) -> None:
        """Display the news."""
        asset = self.market_data.get_asset_from_ticker(self.ticker)
        if asset.name:
            name = remove_punctuation(asset.name.split()[0].title())
        else:
            log.warning(f"Asset not found for ticker: {asset.ticker}")
            name = asset.ticker
        st.markdown(f"## Latest {name} News")

        num_res = int(
            st.sidebar.number_input(
                "Number of news articles",
                min_value=1,
                max_value=10,
                value=3,
                step=1,
            )
        )
        news = self.market_data.get_news((self.ticker,), limit=num_res)
        for n in news:
            if self.ticker in n.symbols:
                st.markdown(f"#### {n.headline}")
                col1, col2 = st.columns(2)
                with col1:
                    st.markdown(f"_{n.author}_")
                with col2:
                    st.write(n.created_at)
                with st.expander("🗞️📰 Full article 👇🏻"):
                    st.write(n.content, unsafe_allow_html=True)

    @staticmethod
    def _display_value_in_container(description: str, value: str | list) -> None:
        """Display the value in an expander container."""
        with st.expander(description.title(), expanded=True):
            if isinstance(value, str):
                st.write(value)
            elif isinstance(value, list):
                for v in value:
                    st.write(v)
            else:
                raise ValueError(f"value must be either a str or a list of str, not {type(value)}")

    def display_optitrader_problem(self) -> None:
        """Display optimization problem."""
        with st.sidebar:
            st.header("Optimization Problem")
            self._display_value_in_container(
                description="investment universe",
                value=self.universe_name.value,
            )
            self._display_value_in_container(
                description="start date",
                value=str(self.start_date.date()),
            )
            self._display_value_in_container(
                description="objectives",
                value=self.objective_names,
            )
            self._display_value_in_container(
                description="constraints",
                value=self.constraint_names,
            )

    def display_alpaca_account_sidebar(self) -> None:
        """Display alpaca account in the page sidebar."""
        with st.sidebar:
            st.header("🦙 Alpaca Account")
            st.write("The default data come from my personal Alpaca trading sandbox API keys.")
            st.markdown(
                """
                [You can find your api keys here](https://alpaca.markets/docs/trading/getting_started/#creating-an-alpaca-account-and-finding-your-api-keys)
                and enter them below to use your account 👇"""
            )
            self.set_api_keys()

    def invest_in_optimal_portfolio(self) -> None:
        """Handle how to invest in the optimal portfolio."""
        if self._opt_ptf:
            bp = self.trader.account.non_marginable_buying_power
            if bp:
                amount = round(
                    st.sidebar.number_input(
                        "Investment amount in USD $",
                        min_value=10.0,
                        max_value=round(number=float(bp), ndigits=2),
                        value=round(number=float(bp) / 5.0, ndigits=1),
                        step=0.1,
                    ),
                    2,
                )
        orders = None
        if st.button(label="INVEST IN PORTFOLIO", disabled=self._opt_ptf is None) and self._opt_ptf:
            with st.spinner("Sending orders..."):
                orders = self.trader.invest_in_portfolio(portfolio=self._opt_ptf, amount=amount)
                st.success("Orders submitted successfully!", icon="✅")
        if self._opt_ptf and orders:
            self._orders_to_st(pd.DataFrame([o.dict() for o in orders]).set_index("symbol"))



"""Portfolio optimization problem module."""
import logging
from enum import Enum
from typing import Any

import cvxpy as cp
import pandas as pd

from optitrader.config import SETTINGS
from optitrader.enums import ConstraintName
from optitrader.enums.optimization import ObjectiveName
from optitrader.optimization.constraints import PortfolioConstraint
from optitrader.optimization.objectives import (
    FinancialsObjectiveFunction,
    ObjectivesMap,
    ObjectiveValue,
    OptimizationVariables,
    PortfolioObjective,
)
from optitrader.portfolio import Portfolio

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class _CVXPYSolver(str, Enum):
    """CVXPY supported solvers."""

    ECOS = "ECOS"
    SCS = "SCS"
    OSQP = "OSQP"


class Solver:
    """Portfolio optimization solver."""

    def __init__(
        self,
        returns: pd.DataFrame,
        objectives: list[PortfolioObjective],
        constraints: list[PortfolioConstraint],
        financials_df: pd.DataFrame | None = None,
    ) -> None:
        if returns.isna().sum().sum():
            raise AssertionError("Passed `returns` contains NaN.")
        self.returns = returns
        if any(isinstance(o, FinancialsObjectiveFunction) for o in objectives):
            assert isinstance(
                financials_df, pd.DataFrame
            ), "You must pass the `financials_df` parameter."
            if financials_df.isna().sum().sum():
                values_per_column = 4
                selected_values = {}
                try:
                    for column in financials_df.columns:
                        non_null_values = (
                            financials_df[column].dropna().iloc[:values_per_column].tolist()
                        )
                        if len(non_null_values) == values_per_column:
                            selected_values[column] = non_null_values
                        else:
                            log.warning(
                                f"{column} with {len(non_null_values)} instead of {values_per_column} dates of financial statements data. Replacing with 0s."
                            )
                            selected_values[column] = [
                                -1e12 for _ in range(values_per_column)
                            ]  # penalize missing data
                    financials_df = pd.DataFrame(selected_values)
                except Exception as e:
                    raise AssertionError("Passed `financials_df` contains NaN.") from e
            financials_df = financials_df.pct_change().iloc[1:].fillna(0).T
            assert not financials_df.empty
        self.financials_df = financials_df
        self.objectives = objectives
        self.constraints = constraints
        self._universe = list(self.returns.columns)
        self._objectives_map = ObjectivesMap(objectives)

    def _get_cvxpy_objectives_and_constraints(
        self, weights_variable: cp.Variable
    ) -> tuple[list[OptimizationVariables], list[cp.Constraint]]:
        """Get portfolio optimization problem."""
        assert (
            self.objectives
        ), "To get a portfolio optimization problem, at least one objective is needed."
        cvxpy_objectives: list[OptimizationVariables] = []
        cvxpy_constraints: list[cp.Constraint] = []
        for obj_fun in self.objectives:
            objective, constr_list = obj_fun.get_objective_and_auxiliary_constraints(
                returns=self.financials_df
                if (
                    isinstance(obj_fun, FinancialsObjectiveFunction)
                    or obj_fun.name == ObjectiveName.FINANCIALS
                )
                else self.returns,
                weights_variable=weights_variable,
            )
            cvxpy_objectives.append(objective)
            cvxpy_constraints.extend(constr_list)
        for constr_fun in self.constraints:
            cvxpy_constraints.extend(
                constr_fun.get_constraints_list(weights_variable=weights_variable)
            )
        return cvxpy_objectives, cvxpy_constraints

    def solve(
        self,
        created_at: pd.Timestamp | None = None,
        weights_tolerance: float | None = SETTINGS.SUM_WEIGHTS_TOLERANCE,
        cvxpy_solver: _CVXPYSolver = _CVXPYSolver.ECOS,
        **kwargs: Any,
    ) -> Portfolio:
        """Solve a portfolio optimization problem.

        Parameters
        ----------
        weights_tolerance
            An optional float, if provided the weights resulting smaller then weights_tolerance
            after an optimization will be set to 0.
        kwargs
            All the supported params of cvxpy.problems.problem.Problem.solve().
        """
        weights_var = cp.Variable(len(self._universe))
        cvxpy_objectives, cvxpy_constraints = self._get_cvxpy_objectives_and_constraints(
            weights_var
        )
        problem = cp.Problem(
            # objective=cp.sum([min_obj for obj in cvxpy_objectives for min_obj in obj.values()]),
            objective=cp.sum([obj.minimize for obj in cvxpy_objectives]),
            constraints=cvxpy_constraints,
        )
        try:
            problem.solve(solver=cvxpy_solver.value, **kwargs)
        except cp.SolverError as se:
            log.warning(se)
            # try with default solver configuration
            problem.solve()
        if problem.status != "optimal":
            raise AssertionError(f"Problem status is not optimal but: {problem.status}")
        weights_series = pd.Series(dict(zip(self._universe, weights_var.value, strict=True)))
        if ConstraintName.SUM_TO_ONE in [c.name for c in self.constraints]:
            assert 1 - weights_series.sum() <= SETTINGS.SUM_WEIGHTS_TOLERANCE
        elif ConstraintName.LONG_ONLY in [c.name for c in self.constraints]:
            assert all(weights_series >= 0)
        if weights_tolerance is not None:
            weights_series[abs(weights_series) < weights_tolerance] = 0.0
        return Portfolio(
            weights=weights_series,
            objective_values=[
                ObjectiveValue(
                    name=cvxpy_obj.name,
                    value=cvxpy_obj.minimize.value,
                    weight=self._objectives_map.get_objective_by_name(cvxpy_obj.name).weight,
                )
                for cvxpy_obj in cvxpy_objectives
            ],
            created_at=created_at,
        )



"""Objectives."""
from abc import ABCMeta, abstractmethod

import cvxpy as cp

from optitrader.enums import ConstraintName


class PortfolioConstraint(metaclass=ABCMeta):
    """Objective function abstract class."""

    def __init__(self, name: ConstraintName) -> None:
        self.name = name

    @abstractmethod
    def get_constraints_list(self, weights_variable: cp.Variable) -> list[cp.Constraint]:
        """Get optimization matrices."""


class UnboundedConstraint(PortfolioConstraint):
    """UnboundedConstraint constraint."""

    def get_constraints_list(self, weights_variable: cp.Variable) -> list[cp.Constraint]:
        """Get no short sell constraint matrices."""
        return []


class BoundedConstraint(PortfolioConstraint):
    """Bounded constraint."""

    def __init__(
        self,
        name: ConstraintName,
        lower_bound: int | None = None,
        upper_bound: int | None = None,
    ) -> None:
        super().__init__(name=name)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def set_bounds(
        self,
        lower_bound: int | None = None,
        upper_bound: int | None = None,
    ) -> None:
        """Set lower_bound."""
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound


class NoShortSellConstraint(UnboundedConstraint):
    """NoShortSell constraint."""

    def __init__(self) -> None:
        super().__init__(name=ConstraintName.LONG_ONLY)

    def get_constraints_list(self, weights_variable: cp.Variable) -> list[cp.Constraint]:
        """Get no short sell constraint matrices."""
        return [weights_variable >= 0]


class SumToOneConstraint(UnboundedConstraint):
    """SumToOne constraint."""

    def __init__(self) -> None:
        super().__init__(name=ConstraintName.SUM_TO_ONE)

    def get_constraints_list(self, weights_variable: cp.Variable) -> list[cp.Constraint]:
        """Get sum to one constraint matrices."""
        return [cp.sum(weights_variable) == 1]


class NumberOfAssetsConstraint(BoundedConstraint):
    """
    NumberOfAssets constraint.

    TODO: See other implementation https://groups.google.com/g/cvxpy/c/l9NetwWXQ-k
    """

    def __init__(
        self,
        lower_bound: int | None = None,
        upper_bound: int | None = None,
    ) -> None:
        super().__init__(name=ConstraintName.NUMER_OF_ASSETS)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def get_constraints_list(self, weights_variable: cp.Variable) -> list[cp.Constraint]:
        """Get sum to one constraint matrices."""
        w_bool = cp.Variable(weights_variable.shape, boolean=True)
        constraints = [weights_variable - w_bool <= 0]
        if self.lower_bound is not None:
            constraints.append(cp.sum(w_bool) >= self.lower_bound)
        if self.upper_bound is not None:
            constraints.append(cp.sum(w_bool) <= self.upper_bound)
        return constraints


class WeightsConstraint(BoundedConstraint):
    """Weights percentage constraint."""

    def __init__(
        self,
        lower_bound: int | None = None,
        upper_bound: int | None = None,
    ) -> None:
        super().__init__(name=ConstraintName.WEIGHTS_PCT)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def get_constraints_list(self, weights_variable: cp.Variable) -> list[cp.Constraint]:
        """Get weights constraints list."""
        _tot = 100
        constraints = []
        if self.lower_bound is not None:
            assert self.lower_bound >= 0, "The lower bound percentage cannot be less than 0."
            assert (
                self.lower_bound <= _tot
            ), f"The lower bound percentage cannot be more than {_tot}."
            constraints.append(weights_variable >= self.lower_bound / _tot)
        if self.upper_bound is not None:
            assert (
                self.upper_bound <= _tot
            ), f"The upper bound percentage cannot be more than {_tot}."
            constraints.append(weights_variable <= self.upper_bound / _tot)
        return constraints


class ConstraintsMap:
    """Constraints map."""

    def __init__(
        self,
        constraints: list[PortfolioConstraint] | None = None,
        constraint_names: list[ConstraintName] | None = None,
    ) -> None:
        self.constraint_mapping: dict[ConstraintName, UnboundedConstraint | BoundedConstraint] = {
            ConstraintName.SUM_TO_ONE: SumToOneConstraint(),
            ConstraintName.LONG_ONLY: NoShortSellConstraint(),
            ConstraintName.NUMER_OF_ASSETS: NumberOfAssetsConstraint(),
            ConstraintName.WEIGHTS_PCT: WeightsConstraint(),
        }
        self.constraints: list[PortfolioConstraint] = (
            constraints or [self.to_constraint(name) for name in constraint_names]
            if constraint_names
            else []
        )

    @property
    def constraints_names(self) -> list[str]:
        """Return the obejctives names."""
        return [o.name for o in self.constraints]

    def to_constraint(
        self,
        name: ConstraintName,
        lower_bound: int | None = None,
        upper_bound: int | None = None,
    ) -> PortfolioConstraint:
        """Get a constraint."""
        if name.is_bounded:
            constr = self.constraint_mapping[name]
            assert isinstance(constr, BoundedConstraint)
            constr.set_bounds(
                lower_bound=lower_bound,
                upper_bound=upper_bound,
            )
            return constr
        return self.constraint_mapping[name]

    def reset_constraint_names(self, constraint_names: list[ConstraintName]) -> None:
        """Reset the constraint names based on the chosen ones in streamlit."""
        self.constraints = [self.to_constraint(name) for name in constraint_names]

    def get_constraint_by_name(
        self,
        name: ConstraintName,
    ) -> PortfolioConstraint:
        """Get an constraint."""
        _objs = [o for o in self.constraints if o.name == name]
        assert len(_objs) > 0, f"Constraint {name} not in the constraints list {self.constraints}."
        return _objs[0]

    def add_constraint(
        self,
        name: ConstraintName,
        lower_bound: int | None = None,
        upper_bound: int | None = None,
    ) -> None:
        """Add an constraint to the map."""
        if name not in self.constraints_names:
            self.constraints.append(
                self.to_constraint(
                    name=name,
                    lower_bound=lower_bound,
                    upper_bound=upper_bound,
                )
            )

    def get_constr_doc(
        self,
        name: ConstraintName,
    ) -> str:
        """Return the constraint docstring."""
        return (
            self.constraint_mapping[name].__doc__
            or f"{self.constraint_mapping[name].__class__.__name__} documentation."
        )

    def set_constraint_bounds(
        self,
        name: ConstraintName,
        lower_bound: int | None = None,
        upper_bound: int | None = None,
    ) -> None:
        """Set the bounds for a bounded constraint."""
        assert name.is_bounded, "This constraint has no bounds."
        assert lower_bound or upper_bound, "You must set either the lower or upper bound."
        if name in self.constraints_names:
            for constr in self.constraints:
                if name == constr.name:
                    assert isinstance(constr, BoundedConstraint)
                    constr.set_bounds(
                        lower_bound=lower_bound,
                        upper_bound=upper_bound,
                    )
        else:
            self.add_constraint(
                name=name,
                lower_bound=lower_bound,
                upper_bound=upper_bound,
            )



"""General utils."""
import re

import pandas as pd


def remove_punctuation(string: str) -> str:
    """Remove punctuation marks in a string."""
    return re.sub(r"[^\w\s]", "", string.replace(".", " "))


def clean_string(string: str) -> str:
    """Clean up a string and return the cleaned string."""
    return string.replace("_", " ").replace("-", " ")


def rearrange_columns_by_zeros(df: pd.DataFrame) -> pd.DataFrame:
    """
    Rearranges the columns of a DataFrame based on the number of zeros in each column.

    Parameters
    ----------
    `df`: (DataFrame) The input DataFrame.

    Returns
    -------
    `DataFrame`: The rearranged DataFrame.
    """
    # Count the number of zeros in each column
    zeros_count = df.eq(0).sum()

    # Sort the columns based on the number of zeros in ascending order
    sorted_columns = zeros_count.sort_values().index

    # Rearrange the columns in the dataframe
    return df[sorted_columns]



"""Update the number of shares in the assets table."""
import logging
import os

from optitrader.market.db.database import MarketDB
from optitrader.market.market_data import MarketData

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


def main(db: MarketDB, md: MarketData) -> None:
    """Run script function."""
    file_path = os.path.relpath(__file__)
    db.create_tables()
    tickers_in_table = db.get_tickers()[500:550]
    shares = md.get_total_number_of_shares(tuple(tickers_in_table))
    df = db.update_number_of_shares(multi_number_of_shares=shares, updated_by=file_path)
    log.debug(df)


if __name__ == "__main__":
    main(db=MarketDB(), md=MarketData(use_db=False))  # pragma: no cover



"""optitrader package."""

from optitrader.main import Optitrader
from optitrader.market import MarketData
from optitrader.portfolio import Portfolio

__all__ = [
    "Optitrader",
    "Portfolio",
    "MarketData",
]



"""Init."""
from optitrader.models.asset import AssetModel
from optitrader.models.optimization import ObjectiveModel, OptimizationRequest, OptimizationResponse

__all__ = [
    "AssetModel",
    "OptimizationRequest",
    "OptimizationResponse",
    "ObjectiveModel",
]



"""Implementation of Alpaca as DataProvider."""

from functools import lru_cache

import pandas as pd
from alpaca.broker import BrokerClient
from alpaca.data import Adjustment, BarSet, StockBarsRequest, StockHistoricalDataClient, TimeFrame
from alpaca.trading import Asset, AssetClass, AssetStatus, GetAssetsRequest, TradingClient

from optitrader.config import SETTINGS
from optitrader.enums import BarsField
from optitrader.market.base_data_provider import BaseDataProvider
from optitrader.market.news import AlpacaNewsAPI, NewsArticle


class AlpacaMarketData(BaseDataProvider):
    """Class to get market data from Alpaca."""

    def __init__(
        self,
        trading_key: str | None = SETTINGS.ALPACA_TRADING_API_KEY,
        trading_secret: str | None = SETTINGS.ALPACA_TRADING_API_SECRET,
        broker_key: str | None = SETTINGS.ALPACA_BROKER_API_KEY,
        broker_secret: str | None = SETTINGS.ALPACA_BROKER_API_SECRET,
    ) -> None:
        super().__init__()
        is_trading = trading_key and trading_secret
        is_broker = broker_key and broker_secret
        assert (
            is_trading or is_broker
        ), "Either Trading API or Broker API keys must be provided to use this service."
        self.__data_client = (
            StockHistoricalDataClient(
                api_key=trading_key,
                secret_key=trading_secret,
            )
            if is_trading
            else StockHistoricalDataClient(
                api_key=broker_key,
                secret_key=broker_secret,
                use_basic_auth=True,
            )
        )
        self.__asset_client: BrokerClient | TradingClient = (
            TradingClient(
                api_key=trading_key,
                secret_key=trading_secret,
            )
            if is_trading
            else BrokerClient(
                api_key=broker_key,
                secret_key=broker_secret,
            )
        )
        self.__news_client: AlpacaNewsAPI | None = (
            AlpacaNewsAPI(
                api_key=broker_key,
                secret_key=broker_secret,
            )
            if is_broker
            else None
        )

    def get_bars(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp | None = None,
        end_date: pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        """
        Get the daily bars dataframe from alpaca-py historical client.

        Parameters
        ----------
        `tickers`: tuple[str, ...]
            A tuple of str representing the tickers.
        `start_date`: pd.Timestamp
            A pd.Timestamp representing start date.
        `end_date`: pd.Timestamp
            A pd.Timestamp representing end date.

        Returns
        -------
        `bars`
            a pd.DataFrame with the bars for the tickers.
        """
        _last_available_date = pd.Timestamp.utcnow() - pd.Timedelta(15, unit="min")
        _first_available_date = pd.Timestamp("2015-12-12").tz_localize(tz="utc")
        # handle first and last available date from Alpaca
        end_date = (
            end_date
            if end_date and end_date.tz_localize(tz="utc") <= _last_available_date
            else _last_available_date
        )
        start_date = (
            start_date
            if start_date and start_date.tz_localize(tz="utc") >= _first_available_date
            else _first_available_date
        )
        bars = self.__data_client.get_stock_bars(
            StockBarsRequest(
                symbol_or_symbols=sorted(tickers),
                start=start_date,
                end=end_date,
                adjustment=Adjustment.ALL,
                timeframe=TimeFrame.Day,
            )
        )
        assert isinstance(bars, BarSet)
        return bars.df

    def get_prices(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
        bars_field: BarsField = BarsField.CLOSE,
    ) -> pd.DataFrame:
        """
        Get the daily bars dataframe from alpaca-py historical client.

        Parameters
        ----------
        `tickers`: tuple[str, ...]
            A tuple of str representing the tickers.
        `start_date`: pd.Timestamp
            A pd.Timestamp representing start date.
        `end_date`: pd.Timestamp
            A pd.Timestamp representing end date.

        Returns
        -------
        `bars`
            a pd.DataFrame with the bars for the tickers.
        """
        bars = self.get_bars(
            tickers=tickers,
            start_date=start_date,
            end_date=end_date,
        )
        bars.reset_index(inplace=True)
        _index_name = "timestamp"
        prices = bars.pivot(index=_index_name, columns="symbol", values=bars_field)
        prices.index = prices.index.strftime("%Y-%m-%d")
        return prices.ffill().bfill()

    def get_alpaca_asset(self, ticker: str) -> Asset:
        """Get alpaca asset by ticker."""
        asset = self.__asset_client.get_asset(symbol_or_asset_id=ticker)
        assert isinstance(asset, Asset)
        return asset

    @lru_cache(maxsize=256)  # noqa: B019
    def get_alpaca_assets(
        self,
        status: AssetStatus = AssetStatus.ACTIVE,
        asset_class: AssetClass = AssetClass.US_EQUITY,
        tradable: bool = True,
        marginable: bool = True,
        fractionable: bool = True,
    ) -> list[Asset]:
        """Get alpaca assets."""
        assets = self.__asset_client.get_all_assets(
            GetAssetsRequest(
                status=status,
                asset_class=asset_class,
            )
        )
        assert isinstance(assets, list)
        return [
            a
            for a in assets
            if a.tradable == tradable
            and a.marginable == marginable
            and a.fractionable == fractionable
        ]

    @lru_cache(maxsize=256)  # noqa: B019
    def get_alpaca_tickers(
        self,
        status: AssetStatus = AssetStatus.ACTIVE,
        asset_class: AssetClass = AssetClass.US_EQUITY,
        tradable: bool = True,
        marginable: bool = True,
        fractionable: bool = True,
    ) -> list[str]:
        """Get alpaca asset tickers."""
        return [
            a.symbol
            for a in self.get_alpaca_assets(
                status=status,
                asset_class=asset_class,
                tradable=tradable,
                marginable=marginable,
                fractionable=fractionable,
            )
        ]

    def get_assets_df(
        self,
        status: AssetStatus = AssetStatus.ACTIVE,
        asset_class: AssetClass = AssetClass.US_EQUITY,
        tradable: bool = True,
        marginable: bool = True,
        fractionable: bool = True,
    ) -> pd.DataFrame:
        """Get alpaca assets dataframe."""
        return pd.DataFrame(
            [
                a.dict()
                for a in self.get_alpaca_assets(
                    status=status,
                    asset_class=asset_class,
                    tradable=tradable,
                    marginable=marginable,
                    fractionable=fractionable,
                )
            ]
        )

    @lru_cache  # noqa: B019
    def get_asset_by_name(self, name: str) -> Asset:
        """Get an asset by its name."""
        assets = self.get_alpaca_assets()
        search_result = [a for a in assets if name in str(a.name)][0]
        assert isinstance(search_result, Asset), f"{name} not found."
        return search_result

    def get_news(
        self,
        tickers: tuple[str, ...],
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        limit: int = 5,
        include_content: bool = True,
        exclude_contentless: bool = True,
    ) -> list[NewsArticle]:
        """Get news articles."""
        assert self.__news_client, "News client is not defined."
        return self.__news_client.get_news(
            tickers=tickers,
            start=start,
            end=end,
            limit=limit,
            include_content=include_content,
            exclude_contentless=exclude_contentless,
        )

    def get_news_df(
        self,
        tickers: tuple[str, ...],
        start: pd.Timestamp | None = None,
        end: pd.Timestamp | None = None,
        limit: int = 5,
        include_content: bool = True,
        exclude_contentless: bool = True,
    ) -> pd.DataFrame:
        """Get news articles in a df."""
        assert self.__news_client, "News client is not defined."
        return self.__news_client.get_news_df(
            tickers=tickers,
            start=start,
            end=end,
            limit=limit,
            include_content=include_content,
            exclude_contentless=exclude_contentless,
        )



"""Module for loading settings and configurations from .env using pydantic."""
from pydantic import BaseSettings


class Settings(BaseSettings):
    """
    Settings and configurations.

    You can provide in the .env file either the Alpaca Trading API or Broker API keys.
    See more here:
    - [Trading API](https://alpaca.markets/docs/trading/getting_started/#creating-an-alpaca-account-and-finding-your-api-keys)
    - [Broker API](https://alpaca.markets/docs/broker/get-started/#api-keys)
    """

    # ALPACA SETTINGS
    ALPACA_TRADING_API_KEY: str | None = None
    ALPACA_TRADING_API_SECRET: str | None = None
    ALPACA_BROKER_API_KEY: str | None = None
    ALPACA_BROKER_API_SECRET: str | None = None

    # FINNHUB
    FINHUB_API_KEY: str | None = None

    # OPTIMIZATION SETTINGS
    # this means that if the portfolio's weights sum to 0.99 instead of 1 is accepted
    SUM_WEIGHTS_TOLERANCE: float = 1e-2

    # DB SETTINGS
    DB_URI_MARKET: str = "sqlite:///market.db"  # prod
    DB_URI_TEST: str = "sqlite:///test.db"  # test

    class Config:
        """Configuration."""

        env_file = ".env"
        env_file_encoding = "utf-8"

    @property
    def is_trading(self) -> bool:
        """If the settings are for the trading keys."""
        return bool(self.ALPACA_TRADING_API_KEY or self.ALPACA_TRADING_API_SECRET)

    @property
    def is_broker(self) -> bool:
        """If the settings are for the broker keys."""
        return bool(self.ALPACA_BROKER_API_KEY or self.ALPACA_BROKER_API_SECRET)


SETTINGS = Settings()



"""Database models."""
from datetime import datetime
from typing import Any

from alpaca.trading import AssetClass, AssetExchange, AssetStatus
from sqlalchemy import Boolean, Column, Date, DateTime, Enum, Integer, String
from sqlalchemy.orm import declarative_base

# metadata_ = MetaData(schema="market")
# Base = declarative_base(metadata=metadata_)
Base = declarative_base()


class Asset(Base):  # type: ignore
    """The storage model for an asset."""

    __tablename__ = "assets"

    _id = Column("id", Integer, primary_key=True, index=True)

    ticker = Column("ticker", String, unique=True, index=True)
    name = Column("name", String)
    sector = Column("sector", String)
    industry = Column("industry", String)
    website = Column("website", String)
    business_summary = Column("business_summary", String)
    number_of_shares = Column("number_of_shares", Integer)
    country = Column("country", String)
    currency = Column("currency", String)
    logo = Column("logo", String)
    ipo = Column("ipo", Date)
    tradable = Column("tradable", Boolean)
    marginable = Column("marginable", Boolean)
    fractionable = Column("fractionable", Boolean)
    status: Column[AssetStatus] = Column("status", Enum(AssetStatus))
    exchange: Column[AssetExchange] = Column("exchange", Enum(AssetExchange))
    asset_class: Column[AssetClass] = Column("asset_class", Enum(AssetClass))

    updated_at = Column("updated_at", DateTime, default=datetime.utcnow)
    updated_by = Column("updated_by", String, default="system")

    def __repr__(self) -> str:
        """Representation of the asset database model."""
        attributes = ", ".join(
            f"{attr}='{value}'" for attr, value in vars(self).items() if attr[0] != "_"
        )
        return f"<Asset({attributes})>"

    def to_dict(self) -> dict[str, Any]:
        """To dictionary."""
        return {k: v for k, v in vars(self).items() if k[0] != "_"}



"""Backtester implementation."""

import pandas as pd
import streamlit as st

from optitrader import Optitrader, Portfolio
from optitrader.enums import RebalanceFrequency
from optitrader.utils import rearrange_columns_by_zeros


class Portfolios:
    """Container of Portfolio objects."""

    def __init__(self, ptfs: list[Portfolio]) -> None:
        """Portfolio init."""
        self.ptfs = ptfs

    def to_df(self) -> pd.DataFrame:
        """To dataframe of weights at each date."""
        _df = pd.DataFrame({p.created_at: p.weights for p in self.ptfs}).T
        _df.index = _df.index.strftime("%Y-%m-%d")
        return rearrange_columns_by_zeros(_df)


class Backtester:
    """Backtester class."""

    def __init__(
        self,
        opt: Optitrader,
        rebal_freq: RebalanceFrequency = RebalanceFrequency.MONTHLY,
        end: pd.Timestamp | None = None,
        start: pd.Timestamp | None = None,
    ) -> None:
        """Initialize new backtester instance."""
        self.opt = opt
        self.rebal_freq = rebal_freq
        self.end = end or pd.Timestamp.today().normalize()
        self.start = start or (self.end - pd.Timedelta(days=365 * 2)).normalize()
        self.ptfs: list[Portfolio] = []

    def get_rebalance_dates(self) -> pd.DatetimeIndex:
        """Return the list of rebalance dates."""
        return pd.date_range(
            start=self.start,
            end=self.end,
            freq=self.rebal_freq.value,
            normalize=True,
        )

    def compute_portfolios(
        self,
        progress_bar: bool = True,
    ) -> list[Portfolio]:
        """Compute portfolio solutions."""
        dates = self.get_rebalance_dates()
        if progress_bar:
            bar = st.progress(0, "Backtest in progress. Please wait.")
        for idx, date in enumerate(dates):
            bar.progress((idx + 1) / len(dates), f"Computing optimal portfolio on {date.date()!s}")
            try:
                self.ptfs.append(self.opt.solve(end_date=date))
            except Exception as e:
                bar.error(e)
                raise e from e
        bar.empty()
        return self.ptfs

    def compute_history_values(self) -> pd.Series:
        """Compute the strategy wealth in the history."""
        rets = self.opt.market_data.get_total_returns(
            tickers=self.opt.investment_universe.tickers,
            start_date=self.start,
            end_date=self.end,
        )
        ptfs = Portfolios(self.ptfs or self.compute_portfolios()).to_df()
        return 1 + (rets * ptfs.reindex(rets.index, method="ffill").fillna(0)).sum(axis=1).cumsum()



"""AssetModel base model."""
from datetime import date
from enum import Enum
from typing import Any

import pandas as pd
from alpaca.trading import AssetClass, AssetExchange, AssetStatus
from pydantic import BaseModel, root_validator

from optitrader.utils import clean_string


class _YahooFinnhubCommon(BaseModel):
    """Model for common fields between yahoo query and finnhub API."""

    name: str | None = None
    industry: str | None = None
    website: str | None = None
    number_of_shares: int | None = None


class YahooAssetModel(_YahooFinnhubCommon):
    """Model to represent asset info from Yahoo query."""

    sector: str | None = None
    business_summary: str | None = None


class FinnhubAssetModel(_YahooFinnhubCommon):
    """Model to represent company_profile2 info from Finhub API."""

    country: str
    currency: str
    logo: str
    ipo: date
    ticker: str


class AssetModel(FinnhubAssetModel, YahooAssetModel):
    """Model to represent an asset."""

    weight_in_ptf: float | None = None
    asset_class: AssetClass
    symbol: str | None = None
    exchange: AssetExchange
    status: AssetStatus
    tradable: bool
    marginable: bool
    fractionable: bool

    @root_validator()
    def validate_ticker_symbol(cls, values: dict[str, Any]) -> dict[str, Any]:  # noqa: N805
        """Validate ticker vs symbol conflict."""
        ticker = values.get("ticker", None)
        symbol = values.get("symbol", None)
        if (ticker and symbol) and ticker != symbol or not symbol:
            values["symbol"] = ticker

        return values

    class Config:
        """Configuration."""

        orm_mode = True

    def to_series(self) -> pd.Series:
        """Cast to series."""
        return pd.Series(
            {
                clean_string(k).title(): clean_string(str(v))
                if not isinstance(v, Enum)
                else clean_string(v.value).title()
                for k, v in self.dict(exclude={"weight_in_ptf", "business_summary", "logo"}).items()
                if k not in {"ipo", "ticker"}
            },
            name="",
        )



"""Config base model."""

from pydantic import BaseModel, Extra


class CustomBaseModel(BaseModel):
    """Custom base model with additional config and settings."""

    class Config:
        """Configuration."""

        extra = Extra.forbid



"""Streamlit explore page."""

import logging

import plotly.express as px
import streamlit as st

from optitrader.app import Page, session
from optitrader.utils import remove_punctuation

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def main() -> None:
    """Run dashboard."""
    page = Page(name="Explore financial data & news")
    page.display_title()
    asset = session.market_data.get_asset(ticker=session.ticker)
    submitted = False
    with st.form(key="search"):
        col1, col2 = st.columns(2)
        with col1:
            name = st.text_input("Search for a stock by name", placeholder=asset.name)
            if name:
                session.set_ticker(session.market_data.get_asset_by_name(name).symbol)
        with col2:
            ticker = st.text_input("Search by ticker", placeholder=session.ticker)
            if ticker:
                session.set_ticker(ticker)

        session.display_tickers()
        submitted = st.form_submit_button("SEARCH 🔍", help="Search for a stock by ticker or name.")
    if submitted:
        asset = session.market_data.get_asset(ticker=session.ticker)
    prices = session.market_data.load_prices(
        tickers=(session.ticker,), start_date=session.start_date
    )[session.ticker]
    if not asset.name:
        log.warning(f"Asset {session.ticker} without name.")
        asset.name = session.ticker

    st.write(
        f"""
        <style>
            .container {{
                display: flex;
                align-items: center;
                margin-bottom: 20px;
            }}

            .logo {{
                border-radius: 50%;
                overflow: hidden;
            }}

            .logo img {{
                width: 50px;
                height: 50px;
            }}

            .title {{
                font-size: 30px;
                font-weight: bold;
                margin-left: 20px;
            }}
        </style>

        <div class="container">
            <div class="logo">
                <img src="{asset.logo}" alt="Logo">
            </div>
            <div class="title">{asset.name.title()}</div>
        </div>
        """,
        unsafe_allow_html=True,
    )
    if asset.business_summary:
        with st.expander("Business summary"):
            st.write(asset.business_summary)
    price = prices[-1]
    last_price = prices[-2]
    prev_market_cap = round(asset.number_of_shares * last_price / 1e6, 2)
    market_cap = round(asset.number_of_shares * price / 1e6, 2)
    col1, col2, col3 = st.columns(3)
    with col1:
        st.metric(
            label="Last Day Closing Price ($)",
            value=round(price, 4),
            delta=round(price - last_price, 4),
        )
    with col2:
        st.metric(
            label="Market Capitalization (Mln $)",
            value=market_cap,
            delta=round(market_cap - prev_market_cap, 2),
        )
    with col3:
        st.metric(label="Main Exchange", value=asset.exchange.value)
    st.plotly_chart(
        px.line(
            data_frame=prices,
            x=prices.index,
            y=prices.values,
            labels={"timestamp": "", "y": ""},
            title="Historical end of day closing prices",
        ),
        use_container_width=True,
    )
    session.display_financials()
    with st.sidebar:
        st.header(remove_punctuation(asset.name.split()[0]).split()[0].title())
        st.dataframe(
            asset.to_series(),
            use_container_width=True,
            height=400,
        )
    session.display_news()



"""Create the assets table."""
import logging
import os

from tqdm import tqdm

from optitrader.market.db.database import MarketDB
from optitrader.market.market_data import MarketData

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def main(db: MarketDB, md: MarketData) -> None:
    """Run script function."""
    file_path = os.path.relpath(__file__)
    db.create_tables()
    univ_tickers = md.get_tradable_tickers()
    tickers_in_table = db.get_tickers()
    tickers = tuple(set(univ_tickers).difference(set(tickers_in_table)))
    chunk_size = 50
    log.info(f"Total to write: {len(tickers)} tickers")
    for chunk in tqdm(range(0, len(tickers), chunk_size)):
        tickers_bucket = tickers[chunk : chunk + chunk_size]
        log.info(f"Getting {len(tickers_bucket)} tickers:")
        asset_models = md.get_assets(tickers=tickers_bucket)
        asset_models = [a for a in asset_models if a.ticker not in tickers_in_table]
        log.info(f"Writing {len(asset_models)} tickers:")
        try:
            db.write_assets(asset_models=asset_models, updated_by=file_path)
        except Exception as exc:
            log.warning(exc)
            db.session.rollback()


if __name__ == "__main__":
    main(db=MarketDB(), md=MarketData(use_db=False))  # pragma: no cover



"""Iterable enumerations."""

from enum import Enum


class IterEnum(str, Enum):
    """An iterable enumeration of string."""

    @classmethod
    def get_values_list(cls) -> list[str]:
        """Get the values in a list."""
        return [member.value for member in cls]

    @classmethod
    def get_names_list(cls) -> list[str]:
        """Get the names in a list."""
        return [member.name for member in cls]

    @classmethod
    def get_index_of_value(cls, value: str) -> int:
        """Get the index of the value in the values list."""
        try:
            return cls.get_values_list().index(value)
        except ValueError as ve:
            raise ValueError(f"Value not found in the Enum {cls.__name__}") from ve



"""Initialization module for a data provider."""

from optitrader.market.investment_universe import InvestmentUniverse
from optitrader.market.market_data import MarketData

__all__ = [
    "MarketData",
    "InvestmentUniverse",
]



"""Module to handle interactions with Finhub API."""

import logging
import time
from functools import lru_cache

import finnhub
import pandas as pd

from optitrader.config import SETTINGS
from optitrader.models.asset import FinnhubAssetModel

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class FinnhubClient(finnhub.Client):
    """Finhub API client."""

    def __init__(self, api_key: str | None = None):
        super().__init__(api_key=api_key or SETTINGS.FINHUB_API_KEY)

    @lru_cache  # noqa: B019
    def get_asset_profile(self, ticker: str) -> FinnhubAssetModel | None:
        """Get the company profile for each ticker."""
        try:
            profile = self.company_profile2(symbol=ticker)
        except finnhub.FinnhubAPIException as api_error:
            log.debug(f"Request for ticker {ticker} sleeping 1 second")
            log.debug(type(api_error))
            time.sleep(1)  # wait time limit reset
            profile = self.company_profile2(symbol=ticker)
        try:
            return FinnhubAssetModel(
                **profile,
                industry=profile["finnhubIndustry"],
                website=profile["weburl"],
                number_of_shares=profile["shareOutstanding"],
                finnhub_name=profile["name"],
            )
        except KeyError as ke:
            log.debug(f"{type(ke)} on {ticker}")
            return None

    @lru_cache  # noqa: B019
    def get_companies_profiles(self, tickers: tuple[str, ...]) -> list[FinnhubAssetModel]:
        """Get the company profile for each ticker."""
        profiles = []
        for i, t in enumerate(sorted(tickers)):
            try:
                profile = self.get_asset_profile(ticker=t)
            # On top of all plan's limit, there is a 30 API calls/ second limit.
            except finnhub.FinnhubAPIException as api_error:
                log.debug(f"Request number {i+1} for ticker {t}")
                log.debug(type(api_error))
                time.sleep(1)  # wait time limit reset
                try:
                    profile = self.get_asset_profile(ticker=t)
                except finnhub.FinnhubAPIException as api_error:
                    log.debug(f"Requests number {i+1} for ticker {t} sleeping 60 seconds")
                    log.debug(type(api_error))
                    # get company profile v2 with 60 rpm
                    time.sleep(60)  # wait time 1 minute for limit reset
                    profile = self.get_asset_profile(ticker=t)
            if profile:
                profiles.append(profile)
        return profiles

    def get_companies_df(self, tickers: tuple[str, ...]) -> pd.DataFrame:
        """Get the dataframe of companies profiles."""
        return pd.DataFrame([a.dict() for a in self.get_companies_profiles(tickers)])



"""Implementation of Yahoo as DataProvider."""

import logging
from functools import lru_cache

import pandas as pd
from yahooquery import Ticker

from optitrader.enums.market import BalanceSheetItem, BarsField, CashFlowItem, IncomeStatementItem
from optitrader.market.base_data_provider import BaseDataProvider
from optitrader.models.asset import YahooAssetModel

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class YahooMarketData(BaseDataProvider):
    """Class to get market data from Yahoo."""

    def __init__(self) -> None:
        super().__init__()
        self.financials = [
            *IncomeStatementItem.get_values_list(),
            *CashFlowItem.get_values_list(),
            *BalanceSheetItem.get_values_list(),
        ]

    def parse_ticker_for_yahoo(self, ticker: str) -> str:
        """Replace a dot with a hyphen for yahoo in ticker."""
        return ticker.replace(".", "-")

    def parse_ticker_from_yahoo(self, ticker: str) -> str:
        """Replace a dot with a hyphen for yahoo in ticker."""
        return ticker.replace("-", ".")

    def parse_tickers_for_yahoo(self, tickers: tuple[str, ...]) -> tuple[str, ...]:
        """Replace a dot with a hyphen for yahoo in tickers."""
        return tuple(self.parse_ticker_for_yahoo(t) for t in tickers)

    @lru_cache  # noqa: B019
    def get_bars(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        """
        Get the daily bars dataframe from alpaca-py historical client.

        Parameters
        ----------
        `tickers`: tuple[str, ...]
            A tuple of str representing the tickers.
        `start_date`: pd.Timestamp
            A pd.Timestamp representing start date.
        `end_date`: pd.Timestamp
            A pd.Timestamp representing end date.

        Returns
        -------
        `bars`
            a pd.DataFrame with the bars for the tickers.
        """
        return Ticker(
            symbols=sorted(self.parse_tickers_for_yahoo(tickers)), asynchronous=True
        ).history(start=start_date, end=end_date, adj_ohlc=True)

    def get_prices(
        self,
        tickers: tuple[str, ...],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
        bars_field: BarsField = BarsField.CLOSE,
    ) -> pd.DataFrame:
        """
        Get the daily bars dataframe from alpaca-py historical client.

        Parameters
        ----------
        `tickers`: tuple[str, ...]
            A tuple of str representing the tickers.
        `start_date`: pd.Timestamp
            A pd.Timestamp representing start date.
        `end_date`: pd.Timestamp
            A pd.Timestamp representing end date.

        Returns
        -------
        `bars`
            a pd.DataFrame with the bars for the tickers.
        """
        bars = self.get_bars(
            tickers=tickers,
            start_date=start_date,
            end_date=end_date,
        )
        bars.reset_index(inplace=True)
        _index_name = "date"
        return bars.pivot(index=_index_name, columns="symbol", values=bars_field)

    def get_yahoo_asset(self, ticker: str, fail_on_yf_error: bool = False) -> YahooAssetModel:
        """Get asset info from yahoo."""
        ticker = self.parse_ticker_for_yahoo(ticker)
        try:
            _ticker = Ticker(ticker)
            _profile = _ticker.asset_profile
            if fail_on_yf_error:
                assert isinstance(_profile, dict), f"Yahoo query returned {_profile}"
            elif _profile is None or isinstance(_profile, str):
                # create empty model with None
                return YahooAssetModel()
            elif isinstance(_profile, dict):
                _profile = _ticker.asset_profile[ticker]
            return YahooAssetModel(
                **_profile,
                business_summary=_profile["longBusinessSummary"],
                number_of_shares=_ticker.key_stats[ticker]["sharesOutstanding"],
            )
        except Exception as exc:
            log.debug(f"{ticker}: {type(exc)}")
            if fail_on_yf_error:
                raise AssertionError(f"Yahoo query returned {_profile}") from exc
            return YahooAssetModel()

    def get_number_of_shares(self, ticker: str) -> int:
        """Get the sharesOutstanding field from yahoo query."""
        _shares = Ticker(self.parse_ticker_for_yahoo(ticker)).key_stats[ticker]
        return int(_shares["sharesOutstanding"]) if isinstance(_shares, dict) else 0

    def get_multi_number_of_shares(self, tickers: tuple[str, ...]) -> pd.Series:
        """Get the sharesOutstanding field from yahoo query."""
        tickers = self.parse_tickers_for_yahoo(tickers)
        y_tickers = Ticker(symbols=sorted(tickers), asynchronous=True, max_workers=10)
        return pd.Series(
            {
                self.parse_ticker_from_yahoo(ticker): int(
                    y_tickers.key_stats[ticker]["sharesOutstanding"]
                )
                if isinstance(y_tickers.key_stats[ticker], dict)
                else 0
                for ticker in tickers
            }
        )

    @lru_cache  # noqa: B019
    def get_financials(self, ticker: str) -> pd.DataFrame:
        """Get financials from yahoo finance."""
        ticker = self.parse_ticker_for_yahoo(ticker)
        fin_df = Ticker(ticker).get_financial_data(
            types=self.financials,
            frequency="q",
            trailing=False,
        )
        assert isinstance(fin_df, pd.DataFrame)
        fin_df = fin_df.reset_index().set_index("asOfDate")
        return fin_df[self.financials]

    @lru_cache  # noqa: B019
    def get_multi_financials_by_item(
        self,
        tickers: tuple[str, ...],
        financial_item: IncomeStatementItem
        | CashFlowItem
        | BalanceSheetItem = IncomeStatementItem.NET_INCOME,
    ) -> pd.DataFrame:
        """Get financials from yahoo finance."""
        tickers = self.parse_tickers_for_yahoo(tickers)
        fin_df = Ticker(tickers).get_financial_data(
            types=[financial_item],
            frequency="q",
            trailing=False,
        )
        assert isinstance(fin_df, pd.DataFrame)
        return fin_df.pivot_table(values=financial_item.value, index="asOfDate", columns="symbol")



"""Streamlit about page."""

from optitrader.app import Page


def main() -> None:
    """Run dashboard."""
    page = Page(name="About optitrader")
    page.display_title_and_description(
        description="""
        optitrader is an open-source Python repository for portfolio optimization and quantitative finance.
        """
    )
    page.display_code_sidebar(with_divider=False)



"""Main module for the optitrader class implementation."""

import pandas as pd

from optitrader.config import SETTINGS
from optitrader.enums import UniverseName
from optitrader.enums.market import BalanceSheetItem, CashFlowItem, IncomeStatementItem
from optitrader.enums.optimization import ObjectiveName
from optitrader.market import InvestmentUniverse, MarketData
from optitrader.optimization.constraints import (
    NoShortSellConstraint,
    NumberOfAssetsConstraint,
    PortfolioConstraint,
    SumToOneConstraint,
    WeightsConstraint,
)
from optitrader.optimization.objectives import PortfolioObjective
from optitrader.optimization.solver import Solver
from optitrader.portfolio import Portfolio


class Optitrader:
    """
    Optitrader main class.

    Attributes
    ----------
    `investment_universe`: InvestmentUniverse
        The investment universe to consider.
    `objectives`: list[PortfolioObjective]
        The objectives of the optimal portfolio.
    `constraints`: list[PortfolioConstraint]
        The constraints of the optimal portfolio.
    `market_data`: MarketData
        The market data instance to get the data from,
        you can pass your own with your API keys and preferred data provider.
    """

    def __init__(
        self,
        objectives: list[PortfolioObjective],
        universe_name: UniverseName | None = None,
        tickers: tuple[str, ...] | None = None,
        constraints: list[PortfolioConstraint] | None = None,
        market_data: MarketData | None = None,
        trading_key: str | None = SETTINGS.ALPACA_TRADING_API_KEY,
        trading_secret: str | None = SETTINGS.ALPACA_TRADING_API_SECRET,
        broker_key: str | None = SETTINGS.ALPACA_BROKER_API_KEY,
        broker_secret: str | None = SETTINGS.ALPACA_BROKER_API_SECRET,
    ) -> None:
        """
        Initialize optitrader instance.

        Parameters
        ----------
        `objectives`: list[PortfolioObjective]
            The objectives of the optimal portfolio.
        `universe_name`: UniverseName | None = None
            The name of investment universe to consider. Mutually exclusive with tickers.
        `tickers`: tuple[str] | None = None
            The tickers investment universe to consider. Mutually exclusive with universe_name.
        `constraints`: list[PortfolioConstraint] | None = None
            The constraints of the optimal portfolio.
        `market_data`: MarketData | None = None
            The market data instance to get the data from,
            you can pass your own with your API keys and preferred data provider.
        """
        assert (
            market_data or (trading_key and trading_secret) or (broker_key and broker_secret)
        ), "You must pass either a MarketData instance, the Trading API keys or Broker API keys."
        self.market_data = market_data or MarketData(
            trading_key=trading_key,
            trading_secret=trading_secret,
            broker_key=broker_key,
            broker_secret=broker_secret,
        )
        self.investment_universe = InvestmentUniverse(tickers=tickers, name=universe_name)
        self.objectives = objectives
        self.constraints = constraints or [SumToOneConstraint(), NoShortSellConstraint()]

    def add_constraint(self, constraint: PortfolioConstraint) -> None:
        """Add a constraint."""
        self.constraints.append(constraint)

    def add_objective(self, objective: PortfolioObjective) -> None:
        """Add an objective."""
        self.objectives.append(objective)

    def solve(
        self,
        start_date: pd.Timestamp | None = None,
        end_date: pd.Timestamp | None = None,
        weights_tolerance: float | None = 0.000001,
        num_assets: int | None = None,
        min_num_assets: int | None = None,
        max_num_assets: int | None = None,
        max_weight_pct: int | None = None,
        min_weight_pct: int | None = None,
        financial_item: IncomeStatementItem
        | CashFlowItem
        | BalanceSheetItem = IncomeStatementItem.NET_INCOME,
    ) -> Portfolio:
        """
        Solve the optimization problem and return the optimal portfolio.

        Parameters
        ----------
        `start_date`: pd.Timestamp
            Starting date of the optimization.
        `end_date`: pd.Timestamp
            Last date for the optimization.
        `weights_tolerance`: float
            Weights less than this tolerance are considered zeros.
        `num_assets`: int
            The number of assets you want in the optimal portfolio.
        `min_num_assets`: int
            The minimum number of assets you want in the optimal portfolio.
        `max_num_assets`: int
            The maximum number of assets you want in the optimal portfolio.
        `max_weight_pct`: int
            The maximum weight for assets you want in the optimal portfolio.
        `min_weight_pct`: int
            The minimum weight for assets you want in the optimal portfolio.

        Returns
        -------
        `opt_ptf`: Portfolio
            The optimal portfolio from the Solver.
        """
        if num_assets:
            assert (
                not min_num_assets
            ), "Cannot provide minimum number of assets if you provided the exact number of assets."
            assert (
                not max_num_assets
            ), "Cannot provide maximum number of assets if you provided the exact number of assets."
            if num_assets == len(self.investment_universe):
                self.add_constraint(
                    WeightsConstraint(
                        lower_bound=1,  # give a minimum of 1 pct weight to each asset
                        upper_bound=max_weight_pct,
                    )
                )
            else:
                self.add_constraint(
                    NumberOfAssetsConstraint(lower_bound=num_assets, upper_bound=num_assets)
                )
        elif min_num_assets or max_num_assets:
            assert (
                not num_assets
            ), "Cannot provide exact number of assets if you provided the minimum or maximum number of assets."
            self.add_constraint(
                NumberOfAssetsConstraint(lower_bound=min_num_assets, upper_bound=max_num_assets)
            )
        elif min_weight_pct or max_weight_pct:
            self.add_constraint(
                WeightsConstraint(
                    lower_bound=min_weight_pct,
                    upper_bound=max_weight_pct,
                )
            )
        opt_ptf = Solver(
            returns=self.market_data.get_total_returns(
                tickers=self.investment_universe.tickers,
                start_date=start_date or end_date - pd.Timedelta(days=365 * 2),
                end_date=end_date,
            ),
            constraints=self.constraints,
            objectives=self.objectives,
            financials_df=self.market_data.get_multi_financials_by_item(
                tickers=self.investment_universe.tickers,
                financial_item=financial_item,
            )
            if ObjectiveName.FINANCIALS in [o.name for o in self.objectives]
            else None,
        ).solve(
            weights_tolerance=weights_tolerance,
            created_at=end_date,
        )
        opt_ptf.set_market_data(self.market_data)
        return opt_ptf



"""Init."""

from optitrader.enums.backtester import RebalanceFrequency
from optitrader.enums.iterable import IterEnum
from optitrader.enums.market import BarsField, DataProvider, UniverseName
from optitrader.enums.optimization import ConstraintName, ObjectiveName

__all__ = [
    "BarsField",
    "DataProvider",
    "UniverseName",
    "IterEnum",
    "ConstraintName",
    "ObjectiveName",
    "RebalanceFrequency",
]



"""Streamlit homepage."""
import streamlit as st

from optitrader.app import Page, session


def main() -> None:
    """Run dashboard."""
    page = Page(name="optitrader dashboard")
    page.display_title_and_description(
        description="""
        In this dashboard you can configure your optimal portfolio using several parameters.\n
        You can select the stocks that the solver should take into consideration when choosing the _investment universe_,
        then you can add and configure _objectives_ and _constraints_ for the optimization problem.\n
        Once you compute the optimal portfolio, analyze the solution obtained and
        head to the _backtester_ page to see how your strategy would've performed historically.\n
        """
    )
    col1, col2 = st.columns(2)
    with col1:
        session.set_universe_name()
        session.set_objective_names()
    with col2:
        session.set_start_date()
        session.set_constraint_names()
    session.run_optimization()
    session.invest_in_optimal_portfolio()
    session.display_solution()
    with st.sidebar:
        st.subheader("⚙️ Configure Optimization")
        session.set_objectives()
        session.set_constraints()
    if session._opt_ptf:
        session.display_news()
    page.display_code_sidebar()



"""Add an asset to the assets table."""
import logging
import os

from optitrader.market.db.database import MarketDB
from optitrader.market.market_data import MarketData

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def main(ticker: str, db: MarketDB, md: MarketData) -> None:
    """Run script function."""
    file_path = os.path.relpath(__file__)
    db.create_tables()
    tickers_in_table = db.get_tickers()
    if ticker not in tickers_in_table:
        asset_model = md.get_asset(ticker)
        log.debug(asset_model)
        db.write_asset(asset_model=asset_model, updated_by=file_path)


if __name__ == "__main__":
    main(ticker="WMT", db=MarketDB(), md=MarketData(use_db=False))  # pragma: no cover



"""News module."""

from datetime import datetime
from functools import lru_cache

from alpaca.common.rest import RESTClient
from pandas import DataFrame, Timestamp
from pydantic import BaseModel

from optitrader.config import SETTINGS


class NewsArticle(BaseModel):
    """The data model for a news article."""

    created_at: datetime
    headline: str
    author: str
    content: str
    symbols: list[str]


class AlpacaNewsAPI(RESTClient):
    """
    Alpaca Rest News API.

    See more:
    - https://github.com/Ale-Cas/optitrader/issues/19
    - https://docs.alpaca.markets/docs/news-api#get-latest-news
    """

    def __init__(
        self,
        api_key: str | None = None,
        secret_key: str | None = None,
        use_basic_auth: bool = True,
    ) -> None:
        """
        Initialize a News Data Client.

        Args:
            api_key (Optional[str], optional): Alpaca API key. Defaults to None.
            secret_key (Optional[str], optional): Alpaca API secret key. Defaults to None.
            oauth_token (Optional[str]): The oauth token if authenticating via OAuth. Defaults to None.
            use_basic_auth (bool, optional): If true, API requests will use basic authorization headers.
        """
        assert SETTINGS.is_broker, "You must set the Broker API keys to use this API."
        super().__init__(
            api_key=api_key or SETTINGS.ALPACA_BROKER_API_KEY,
            secret_key=secret_key or SETTINGS.ALPACA_BROKER_API_SECRET,
            use_basic_auth=use_basic_auth,
            api_version="v1beta1",
            base_url="https://data.sandbox.alpaca.markets",
            sandbox=True,
            raw_data=True,  # not mapped by alpaca-py
        )

    @lru_cache  # noqa: B019
    def get_news(
        self,
        tickers: tuple[str, ...],
        start: datetime | Timestamp | None = None,
        end: datetime | Timestamp | None = None,
        limit: int = 5,
        include_content: bool = True,
        exclude_contentless: bool = True,
    ) -> list[NewsArticle]:
        """Get news articles."""
        result = self.get(
            path="/news",
            data={
                "symbols": ",".join(tickers),
                "start": start.isoformat("T") + "Z" if start else None,
                "end": end.isoformat("T") + "Z" if end else None,
                "limit": limit,
                "include_content": include_content,
                "exclude_contentless": exclude_contentless,
            },
        )
        assert isinstance(result, dict), f"Result is not a dict but {type(result)}"
        return [NewsArticle(**res) for res in result["news"]]

    def get_news_df(
        self,
        tickers: tuple[str, ...],
        start: datetime | Timestamp | None = None,
        end: datetime | Timestamp | None = None,
        limit: int = 5,
        include_content: bool = True,
        exclude_contentless: bool = True,
    ) -> DataFrame:
        """Get news articles."""
        return DataFrame(
            n.dict()
            for n in self.get_news(
                tickers=tickers,
                start=start,
                end=end,
                limit=limit,
                include_content=include_content,
                exclude_contentless=exclude_contentless,
            )
        ).set_index("created_at")



"""Portfolio module."""

import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from alpaca.trading import OrderRequest, OrderSide, OrderType, TimeInForce
from typeguard import typechecked

from optitrader.config import SETTINGS
from optitrader.market import MarketData
from optitrader.models import AssetModel
from optitrader.optimization.objectives import ObjectiveValue


class Portfolio:
    """Portfolio class."""

    @typechecked
    def __init__(
        self,
        weights: pd.Series | dict[str, float],
        objective_values: list[ObjectiveValue] | None = None,
        market_data: MarketData | None = None,
        created_at: pd.Timestamp | None = None,
    ) -> None:
        weights = weights if isinstance(weights, pd.Series) else pd.Series(weights)
        if not weights.empty:
            _wsum = weights.values.sum()
            if _wsum:
                assert (
                    1 - _wsum <= SETTINGS.SUM_WEIGHTS_TOLERANCE
                ), f"The sum of weights has to be 1 not {_wsum}."
        self.weights = pd.Series(weights)
        self.objective_values = objective_values or []
        self.market_data = market_data
        self.created_at = created_at or pd.Timestamp.utcnow()

    def __repr__(self) -> str:
        """Object representation."""
        if self.objective_values:
            objectives_dict = {o.name.value: o.value for o in self.objective_values}
            return f"{self.__class__.__name__}(weights={self.get_non_zero_weights().to_dict()}, objective_values={objectives_dict})"
        return f"{self.__class__.__name__}(weights={self.get_non_zero_weights().to_dict()}"

    def get_non_zero_weights(self, round_to_decimal: int | None = 5) -> pd.Series:
        """Non zero weights."""
        non_zero = self.weights[self.weights != 0]
        return non_zero.round(round_to_decimal) if round_to_decimal else non_zero

    def get_tickers(self, only_non_zero: bool = True) -> tuple[str, ...]:
        """Get the tickers in portfolio."""
        weights = self.get_non_zero_weights() if only_non_zero else self.weights
        return tuple(weights.keys())

    @typechecked
    def set_market_data(self, market_data: MarketData) -> None:
        """Set the market data."""
        self.market_data = market_data

    def get_assets_in_portfolio(self, only_non_zero: bool = True) -> list[AssetModel]:
        """Return the assets in the portfolio."""
        assert isinstance(
            self.market_data, MarketData
        ), "You must set the market data to get the assets info."
        weights = self.get_non_zero_weights() if only_non_zero else self.weights
        assets = self.market_data.get_assets(tickers=tuple(weights.keys()))
        assets_in_ptf = []
        for asset in assets:
            asset.weight_in_ptf = weights.get(asset.ticker)
            assets_in_ptf.append(asset)
        return assets_in_ptf

    def get_assets_df(self) -> pd.DataFrame:
        """Return the assets in the portfolio."""
        assert isinstance(
            self.market_data, MarketData
        ), "You must set the market data to get the assets info."
        weights = self.get_non_zero_weights()
        weights.name = "weight_in_ptf"
        assets = self.market_data.get_assets_df(tuple(weights.keys()))
        assets.set_index("ticker", inplace=True)
        return pd.concat([assets, weights], axis=1)

    def get_holdings_df(self) -> pd.DataFrame:
        """Return holdings info df."""
        df = self.get_assets_df()
        if not df.empty:
            return df.sort_values(by="weight_in_ptf", ascending=False)
        return df

    def get_history(
        self,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp | None = None,
    ) -> pd.Series:
        """
        Get the portfolio wealth history.

        Parameters
        ----------
        `start_date`: pd.Timestamp
            The starting date.
        `end_date`: pd.Timestamp
            The ending date. Defaults today.

        Returns
        -------
        `history`: pd.Series
            A timeseries with the portfolio value at each date.
        """
        assert isinstance(
            self.market_data, MarketData
        ), "You must set the market data to get the assets info."
        rets = self.market_data.get_total_returns(
            tickers=self.get_tickers(),
            start_date=start_date,
            end_date=end_date,
        )
        return 1 + (rets * self.get_non_zero_weights()).sum(axis=1, skipna=True).cumsum()

    def pie_plot(self, title: str = "Portfolio Allocation") -> go.Figure:
        """
        Display a pie plot of the weights.

        Parameters
        ----------
        `title`: str
            The title of the plot.
        """
        weights = self.get_non_zero_weights()
        return px.pie(
            data_frame=weights,
            names=weights.keys(),
            values=weights.values,
            title=title,
        )

    def history_plot(
        self,
        start_date: pd.Timestamp | None = None,
        end_date: pd.Timestamp | None = None,
        title: str = "Portfolio value from start date to today",
    ) -> go.Figure:
        """
        Display a line plot of the portfolio historical values.

        Parameters
        ----------
        `start_date`: pd.Timestamp
            The starting date. Defaults to a year ago from today.
        `end_date`: pd.Timestamp
            The ending date. Defaults to today.
        `title`: str
            The title of the plot.
        """
        end_date = end_date or pd.Timestamp.today()
        start_date = start_date or end_date - pd.Timedelta(days=365)
        history = (
            self.get_history(start_date=start_date, end_date=end_date)
            if len(self.get_tickers()) > 0
            else pd.Series()
        )
        return px.line(
            data_frame=history,
            x=history.index,
            y=history.values,
            labels={"timestamp": "", "y": ""},
            title=title,
        )

    def to_orders_list(self, amount: float) -> list[OrderRequest]:
        """Make a list of order requests from the portfolio weights."""
        return [
            OrderRequest(
                symbol=ticker,
                notional=round(weight * amount, 2),
                side=OrderSide.BUY,
                type=OrderType.MARKET,
                time_in_force=TimeInForce.DAY,
            )
            for ticker, weight in self.get_non_zero_weights().sort_values(ascending=False).items()
            if round(weight * amount, 2) >= 1
        ]
